from __future__ import print_function
import weakref
import os
import traceback
import wx
import wx.grid
from ifigure.utils.eval_node import EvalNode
from ifigure.utils.cbook import ImageFiles, Write2Main, BuildPopUpMenu
import ifigure.widgets.dialog as dialog

from ifigure.widgets.var_viewerg2 import _PropertyGrid
from ifigure.utils.wx3to4 import isWX3, GridTableBase
# class _PropertyGrid( wx.grid.Grid ):
#     def __init__(self, *args, **kargs):
#         wx.grid.Grid.__init__(self, *args, **kargs)
#         self.Bind(wx.EVT_SIZE, self.onSizeEvent)
#     def onSizeEvent(self, evt):
#         self.ForceRefresh()
#     def SetTable( self, object, *attributes ):
#         self.tableRef = weakref.ref( object )
#         return wx.grid.Grid.SetTable(self, object, *attributes )
#     def GetTable( self ):
#         return self.tableRef()


class ShellVarViewerDropTarget(wx.TextDropTarget):
    # obj is proj_viewer
    def __init__(self, obj):
        wx.TextDropTarget.__init__(self)
        self.obj = obj

    def OnDropText(self, x, y, indata):
        print(type(indata))

        # this is ad-hoc....!
        app = self.obj.GetTopLevelParent()
        indata = app._text_clip
        app._text_clip = ''

        data = '='+str(indata)
        try:
            print(str(indata))
            obj = EvalNode(str(indata))
            # print obj
            if obj.isTreeDict():
                data = data + '[:]'
        except Exception:
            pass
        pv = self.obj
        vv = pv.varviewer

        row = vv.grid.YToRow(y + vv.grid.GetViewStart()[1] *
                             vv.grid.GetScrollPixelsPerUnit()[1])
        if row == -1:
            return
        #        row=vv.grid.YToRow(y)
        gt = vv.grid.GetTable()
        treedict = gt.GetTreeDict()
        if treedict is None:
            return
        if row == -1:
            name = 'link'
            p = 1
            name1 = name+str(p)
            while treedict.getvar(name1) is not None:
                p = 1+p
                name1 = name+str(p)
            treedict.setvar(name1, data)

        else:
            name = (treedict.get_varlist())[row]
            treedict.setvar(name, data)

        vv.fill_list(treedict)
        item = pv.dt1._prev_item
        if pv.dt1._prev_item is not None:
            pv.tree.SetItemDropHighlight(
                pv.dt1._prev_item, False)

        pv.tree.UnselectAll()
        if item is not None:
            pv.tree.SelectItem(item)


class ShellVarViewerPopUp(wx.Menu):
    def __init__(self, parent):
        super(ShellVarViewerPopUp, self).__init__()
        self.parent = parent
        labels = self._cb_label()
        menus = []
        for i, l in enumerate(labels):
            def func(evt, parent=parent, idx=i):
                parent._selection[idx] = not parent._selection[idx]
                parent.update_table()
            menus.append((l, func, None))
        gt = self.parent.grid.GetTable()
        txt = ['*', '*']
        if gt._sort == 0:
            txt[0] = '^'
        elif gt._sort == 1:
            txt[1] = '^'
        menus.extend([('+Sort', None, None),
                      (txt[0] + 'Up',   self.onSortUp, None),
                      (txt[1] + 'Down', self.onSortDown, None),
                      ('!',    None, None)])
        BuildPopUpMenu(self, menus)

    def _cb_label(self):
        names = ['variables', 'func', 'module']
        onoffs = ['hide ' if x else 'show ' for x in self.parent._selection]
        return [onoff+name for onoff, name in zip(onoffs, names)]

    def onSortUp(self, e):
        gt = self.parent.grid.GetTable()
        if gt._sort == 0:
            gt._sort = -1
        else:
            gt._sort = 0
        self.parent.Refresh()

    def onSortDown(self, e):
        gt = self.parent.grid.GetTable()
        if gt._sort == 1:
            gt._sort = -1
        else:
            gt._sort = 1
        self.parent.Refresh()


class ShellVarViewerGridTable(GridTableBase):
    def __init__(self, obj, grid):
        GridTableBase.__init__(self)
        self._obj = []
        self._grid = grid
        self._sort = -1
        self.currentRows = 0
        self.currentCols = self.GetNumberCols()

    def getGrid(self):
        return self._grid

    def set_variable_list(self, obj):
        self._obj = obj
        self.ResetView()

    def GetNumberRows(self):
        """Return the number of rows in the grid"""
        return len(self._obj)

    def GetNumberCols(self):
        """Return the number of columns in the grid"""
        return 3

    def IsEmptyCell(self, row, col):
        """Return True if the cell is empty"""
        return False

    def GetTypeName(self, row, col):
        """Return the name of the data type of the value in the cell"""
        # return None
        return 'string'

    def GetValue(self, row, col):
        """Return the value of a cell"""
        if row >= len(self._obj):
            return ''
        name = self.get_row_name(row)
        for x in self._obj:
            if x[0] == name:
                return x[col+1]
#        return self._obj[row][col+1]

    def SetValue(self, row, col, value):
        """Set the value of a cell"""
        pass

    def GetColLabelValue(self, col):
        if col == 0:
            return 'type'
        if col == 1:
            return 'value'
        if col == 2:
            return 'shape'

    def GetRowLabelValue(self, row):
        if row >= len(self._obj):
            return ''
        return self.get_row_name(row)

    def get_row_name(self, row):
        #        print self._obj
        #        if obj is None: return ''
        names = [x[0] for x in self._obj]
        if row >= len(names):
            return ''
        if self._sort == -1:
            return names[row]
        elif self._sort == 0:
            return sorted(names)[row]
        elif self._sort == 1:
            return [x for x in reversed(sorted(names))][row]

    def GetAttr(self, row, col, someExtraParameter):
        attr = wx.grid.GridCellAttr()
        if (col == 1 or col == 0):
            attr.SetReadOnly(1)
            return attr
        return None

    def ResetView(self):
        """Trim/extend the control's rows and update all values"""
        grid = self.getGrid()
        grid.BeginBatch()
        for current, new, delmsg, addmsg in [
            (self.currentRows, self.GetNumberRows(
            ), wx.grid.GRIDTABLE_NOTIFY_ROWS_DELETED, wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED),
            (self.currentCols, self.GetNumberCols(
            ), wx.grid.GRIDTABLE_NOTIFY_COLS_DELETED, wx.grid.GRIDTABLE_NOTIFY_COLS_APPENDED),
        ]:
            if new < current:
                msg = wx.grid.GridTableMessage(self, delmsg, new, current-new)
                # for i in range(current-new):
                grid.ProcessTableMessage(msg)
                self.currentRows = new
            elif new > current:
                msg = wx.grid.GridTableMessage(self, addmsg,  new-current)
                grid.ProcessTableMessage(msg)
                self.currentRows = new
        self.UpdateValues()
        if len(self._obj) == 0:
            ml = 8
        else:
            ml = max([len(x[0]) for x in self._obj])
            grid.SetRowLabelSize(ml*10)  # 10 is an ad-hoc number
        grid.EndBatch()

        # print 'current row', self.currentRows

        # The scroll bars aren't resized (at least on windows)
        # Jiggling the size of the window rescales the scrollbars
        w, h = grid.GetSize()
        grid.SetSize((w+1, h))
        grid.SetSize((w, h))

        grid.ForceRefresh()

    def fit_col_width(self):
        grid = self.getGrid()
        w, h = grid.GetClientSize()
        d = w - grid.GetRowLabelSize()-grid.GetColSize(1) - grid.GetColSize(0)
        grid.SetColSize(2, max([d, 80]))

    def UpdateValues(self):
        """Update all displayed values"""
        msg = wx.grid.GridTableMessage(self,
                                       wx.grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.getGrid().ProcessTableMessage(msg)
        self.getGrid().ClearSelection()

    def IssueDoDragDrop(self, row):
        if len(self._obj) == 0:
            return ""
        text = self.get_row_name(row)

        app = self._grid.GetTopLevelParent()
        app._text_clip = text
        if not wx.TheClipboard.IsOpened():
            wx.TheClipboard.Open()
            data = wx.TextDataObject()
            data.SetText(text)
            wx.TheClipboard.SetData(data)
            wx.TheClipboard.Close()

        tdo = wx.TextDataObject(text)
        src = self._grid.GetGridRowLabelWindow()
        tds = wx.DropSource(src)
        tds.SetData(tdo)
        tds.DoDragDrop(True)


class ShellVarViewer(wx.Panel):

    def __init__(self, parent):
        wx.Panel.__init__(self, parent)
        #        self.st1 = wx.StaticText(self, label=' = ')
        #        self.ct1 = wx.TextCtrl(self, wx.ID_ANY, value="", style=wx.TE_PROCESS_ENTER)

        self._selection = [True, False, False]  # vars, func, module, class
        # sizer0=wx.BoxSizer(wx.HORIZONTAL)
        # sizer0.Add(self.st1, 0, wx.ALL, 1)
        # sizer0.Add(self.ct1, 0, wx.ALL, 1)
        self.grid = _PropertyGrid(self)
#        self.grid = wx.grid.Grid(self)
        self.grid.CreateGrid(3, 0)
        self.grid.SetTable(ShellVarViewerGridTable(None, self.grid))
        #   self.ct1.SetDropTarget(wx.TextDropTarget())

        bottom = wx.BoxSizer(wx.HORIZONTAL)
        modes = ['copy', 'paste', 'trash']
        for mode in modes:
            if mode == 'text':
                self.tc_path = TextCtrlPath(self, wx.ID_ANY, '')
                self.tc_path.pv = self
#              self.tc_path.Bind(wx.EVT_KEY_DOWN, self.onKeyPressed, self.tc_path)
                bottom.Add(self.tc_path, 1, wx.EXPAND | wx.ALL, 1)
            else:
                from ifigure.ifigure_config import icondir
                path = os.path.join(icondir, '16x16', mode+'.png')
                image = wx.Image(path, wx.BITMAP_TYPE_PNG).ConvertToBitmap()
                bt = wx.BitmapButton(self, -1, image)
                bt.SetToolTip(wx.ToolTip(mode))
                if mode == 'trash':
                    bottom.AddStretchSpacer()
                    bottom.Add(bt, 0)
                else:
                    bottom.Add(bt, 0)

                def func(e, mode=mode): return self.onButton(e, mode)
                self.Bind(wx.EVT_BUTTON, func, bt)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        #   sizer.Add(sizer0, 0, wx.EXPAND, 0)
        sizer.Add(self.grid, 1, wx.EXPAND, 0)
        sizer.Add(bottom, 0, wx.EXPAND)
        #   self.Bind(wx.grid.EVT_GRID_CELL_CHANGE, self.onEdit)
        #   self.ct1.Bind(wx.EVT_TEXT_ENTER, self.onEditValue)

        rowWindow = self.grid.GetGridRowLabelWindow()
        colWindow = self.grid.GetGridColLabelWindow()
        self._potentialDrag = False
        self._startrow = None  # param for drag'n'drop
        self._startrow2 = -1  # param for popup menu
        rowWindow.Bind(wx.EVT_LEFT_DOWN, self.OnDragGridLeftDown)
        rowWindow.Bind(wx.EVT_LEFT_UP, self.OnDragGridLeftUp)
        rowWindow.Bind(wx.EVT_RIGHT_UP, self.OnRightRelease)
        rowWindow.Bind(wx.EVT_RIGHT_DOWN, self.OnRightPress)

        colWindow.Bind(wx.EVT_RIGHT_UP, self.OnRightRelease)
        colWindow.Bind(wx.EVT_RIGHT_DOWN, self.OnRightPress)

        self.Bind(wx.grid.EVT_GRID_COL_SIZE, self.onGridSize)
        self.grid.Bind(wx.EVT_SIZE, self.onGridSize)

    #   self.Bind(wx.grid.EVT_GRID_BEGIN_DRAG, self.OnBeginDrag)

    def onButton(self, evt, mode):
        from ifigure.ifigure_config import vv_scratch
        import ifigure.utils.pickle_wrapper as pickle

        idx = self.grid.GetSelectedRows()
        if len(idx) == 0 and mode != 'paste':
            return
        gt = self.grid.GetTable()
        obj = gt._obj

        if mode == 'copy':
            names = [str(self.grid.GetRowLabelValue(d)) for d in idx]
            self.GetTopLevelParent().set_status_text(
                'Copy ' + ', '.join(names), timeout=3000)
            try:
                fid = open(vv_scratch, 'wb')
                data = {n: self.shell.get_shellvar(n) for n in names}
                pickle.dump(data, fid)
                fid.close()
            except:
                dialog.showtraceback(parent=self,
                                     txt='Failed to copy',
                                     title='Failed to copy',
                                     traceback=traceback.format_exc())
        elif mode == 'paste':
            fid = open(vv_scratch, 'rb')
            data = pickle.load(fid)
            fid.close()
            for key in data:
                if self.shell.has_shellvar(key):
                    dlg = wx.MessageDialog(None,
                                           'Do you want to overwrite '+key + '?',
                                           'Variable already existws',
                                           wx.OK | wx.CANCEL)
                    ret = dlg.ShowModal()
                    dlg.Destroy()
                    if ret != wx.ID_OK:
                        continue
                self.GetTopLevelParent().set_status_text('Paste tree variable', timeout=3000)
                self.shell.set_shellvar(key, data[key])
            self.update()
        elif mode == 'trash':
            names = [str(self.grid.GetRowLabelValue(d)) for d in idx]
            dlg = wx.MessageDialog(None,
                                   'Do you want to delete ' +
                                   ', '.join(names) + '?',
                                   'Deleting variable',
                                   wx.OK | wx.CANCEL)
            ret = dlg.ShowModal()
            dlg.Destroy()
            if ret == wx.ID_OK:
                for n in names:
                    self.shell.del_shellvar(n)
                self.GetTopLevelParent().set_status_text(
                    'Deltete '+', '.join(names), timeout=3000)
                wx.CallAfter(self.update)

    def onGridSize(self, evt):
        self.grid.GetTable().fit_col_width()
        self.grid.ForceRefresh()

    def update(self, shell=None):
        if shell is not None:
            self.shell = shell
        self._varlist = self.shell.list_locals()
        self.update_table()
        #        self.ct1.SetValue('')

    def update_table(self):
        if self._selection[0]:
            l1 = [v for v in self._varlist if (
                v[1] != 'function' and v[1] != 'module' and v[1] != 'class')]
        else:
            l1 = []
        if self._selection[1]:
            l1.extend([v for v in self._varlist if v[1] == 'function'])
        if self._selection[2]:
            l1.extend([v for v in self._varlist if v[1] == 'module'])
        self.grid.GetTable().set_variable_list(l1)

    def SetDropTarget(self, tg):
        self.grid.GetGridWindow().SetDropTarget(tg)
#       uncomment this makes crush on MacOSX
#        self.grid.GetGridRowLabelWindow().SetDropTarget(tg)

    def OnDragGridLeftDown(self, e):
        # this is to convert y to row. need to adjust
        # scroll.....
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
        if row == -1:
            return
        self._startrow = row
        self._potentialDrag = True
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Bind(wx.EVT_MOTION, self.OnDragGridMotion)

        e.Skip()

    def OnDragGridLeftUp(self, e):
        """We are not dragging anymore, so unset the potentialDrag flag"""
        self._potentialDrag = False
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Unbind(wx.EVT_MOTION)

        e.Skip()

    def OnRightPress(self, e):
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
#        row=self.grid.YToRow(e.GetY())
        self._startrow2 = row

    def OnRightRelease(self, e):
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
#        row=self.grid.YToRow(e.GetY())
        if self._startrow2 != row:
            # press and release happend at different row
            self._startrow2 = row
            return

        m = ShellVarViewerPopUp(self)
        self.PopupMenu(m,
                       e.GetPosition())
        m.Destroy()

    def OnDragGridMotion(self, e):
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Unbind(wx.EVT_MOTION)
        self._potentialDrag = False
        if self._startrow is None:
            return

        gt = self.grid.GetTable()
        gt.IssueDoDragDrop(self._startrow)

        e.Skip()

    def setdroptarget(self, proj_viewer):
        dt2 = ShellVarViewerDropTarget(proj_viewer)
        self.SetDropTarget(dt2)

    def Copy(self, e=None):
        idx = self.grid.GetSelectedRows()
        if len(idx) == 0:
            return
        text = str(self.grid.GetRowLabelValue(idx[0]))

        if not wx.TheClipboard.IsOpened():
            wx.TheClipboard.Open()
            data = wx.TextDataObject()
            data.SetText(text)
            wx.TheClipboard.SetData(data)
            wx.TheClipboard.Close()

    def Paste(self, e=None):
        '''
        can not paste to shellvar viewer
        '''
        pass



import numpy as np
import ifigure

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('GL_COMPOUND')


class GLCompound(object):
    def isCompound(self):
        if (self.hasvar('array_idx') and
                self._var['array_idx'] is not None):
            return True
        return False

    @property
    def hidden_component(self):
        if not hasattr(self, '_hidden_component'):
            self._hidden_component = []
        return self._hidden_component

    @property
    def shown_component(self):
        h = self._hidden_component
        array_idx = np.unique(self.getvar('array_idx'))
        idx = array_idx[np.in1d(array_idx, h, invert=True)]

        return list(idx)

    def hide_component(self, idx, inverse=False):
        '''
        idx = list
        '''
        if not self.isCompound():
            return

        if inverse:
            array_idx = np.unique(self.getvar('array_idx'))
            idx = array_idx[np.in1d(array_idx, idx, invert=True)]
            idx = list(idx)

        self._hidden_component = idx
        if len(self._artists) == 0:
            return

        a = self._artists[0]
        array_idx = self.getvar('array_idx')
        flat_mode = array_idx.shape != a._gl_array_idx.shape

        if self.hasvar('idxset'):
            idxset = self.getvar('idxset')

            if flat_mode:
                mask = np.logical_not(
                    np.in1d(
                        np.abs(
                            a._gl_array_idx),
                        self._hidden_component))
                new_idxset = np.arange(len(a._gl_array_idx), dtype=int)
                new_idxset = new_idxset[mask].reshape(-1, idxset.shape[1])

            else:
                mask = np.in1d(array_idx, self._hidden_component)
                mask2 = np.logical_not(np.any(mask[idxset], axis=1))
                new_idxset = idxset[mask2]

            a.update_idxset(new_idxset)
            self.setSelectedIndex([])

            if self.hasvar('edge_idx') and self.getvar('edge_idx') is not None:
                idxset = self.getvar('edge_idx')
                mask2 = np.logical_not(np.any(mask[idxset], axis=1))
                a.update_edge_idxset(idxset[mask2])
        else:
            assert False, "hide_component is not supported for non-indexed artist"

    def get_subset(self, component=None):
        '''
        v, idx, cdata = gl_compound::get_subset(component = None)
        v, idx, cdata, edge_idx = gl_compound::get_subset(component = None)

        return the vertex information of components
        if component is None, it returns visible components
        '''
        if not self.hasvar('idxset'):
            return
        if component is None:
            component = self.shown_component
        array_idx = self.getvar('array_idx')
        idxset = self.getvar('idxset')

        mask = np.array([ii in component for ii in array_idx],
                        copy=None)
        mask2 = np.array([all(mask[iv]) for iv in idxset], copy=None)

        s = idxset[mask2]
        ii, arr = np.unique(s.flatten(), return_inverse=True)
        idx = arr.reshape(s.shape)
        v = self.getvar('v')[ii]
        if not self.hasvar('cdata'):
            cdata = self.hasvar('cdata')[idx]
        else:
            cdata = None

        if self.hasvar('edge_idx') and self.getvar('edge_idx') is not None:
            idxset = self.getvar('edge_idx')
            mask2 = np.array([all(mask[iv]) for iv in idxset], copy=None)
            s2 = idxset[mask2]
            mapper = {x: k for k, x in enumerate(ii)}
            ss2 = np.array([mapper[x] for x in s2.flatten()])
            s2 = ss2.reshape(s2.shape)
            return v, idx, cdata, s2

        return v, idx, cdata

    def isSelected(self):
        # this is called to check if click hit this object
        return (len(self._artists[0]._gl_hit_array_id_new) > 0)

    def getSelectedIndex(self):
        return self._artists[0]._gl_hit_array_id

    def setSelectedIndex(self, ll):
        array_idx = self.getvar('array_idx')
        if array_idx is None:
            return
        array_idx = array_idx.copy()
        mask = np.isin(array_idx, np.array(ll, copy=None))
        array_idx[mask] *= -1
        for a in self._artists:
            a._gl_hit_array_id = list(ll)
            if a._gl_array_idx.shape == array_idx.shape:
                a._gl_array_idx = array_idx
            else:
                array_idx = np.abs(a._gl_array_idx)
                mask = np.isin(array_idx, np.array(ll, copy=None))
                array_idx[mask] *= -1
                a._gl_array_idx = array_idx
            a._update_a = True

    def addSelectedIndex(self, ll):
        ll = list(ll) + self.getSelectedIndex()
        ll = list(np.unique(ll))
        self.setSelectedIndex(ll)

    def set_pickmask(self, value):
        '''
        mask = True :: not pickable
        '''
        self._pickmask = value
        for a in self._artists:
            a._gl_pickable = not value

    def set_gl_hl_use_array_idx(self, value):
        self.setp('_hl_use_array_idx', value)
        for a in self._artists:
            a.set_gl_hl_use_array_idx(value)

    def get_gl_hl_use_array_idx(self):
        return self.getp('_hl_use_array_idx') == True

    def onPickComponent(self, evt):
        self.set_gl_hl_use_array_idx(True)

    def onUnpickComponent(self, evt):
        self.set_gl_hl_use_array_idx(False)

    def onHidePickedComponent(self, evt):
        idx = self.getSelectedIndex()
        hidx = self.hidden_component
        self.hide_component(idx + hidx)
        self.set_bmp_update(False)
        evt.GetEventObject().unselect_all()
        ifigure.events.SendPVDrawRequest(self, w=evt.GetEventObject(),
                                         wait_idle=True, refresh_hl=False)

    def onShowPickedOnly(self, evt):
        idx = self.getSelectedIndex()
        self.hide_component(idx, inverse=True)
        self.set_bmp_update(False)
        ifigure.events.SendPVDrawRequest(self, w=evt.GetEventObject(),
                                         wait_idle=True, refresh_hl=False)

    def onShowAllComponent(self, evt):
        self.hide_component([])
        self.setSelectedIndex([])
        self.set_bmp_update(False)
        self.set_bmp_update(False)
        ifigure.events.SendPVDrawRequest(self, w=evt.GetEventObject(),
                                         wait_idle=True, refresh_hl=False)

    def canvas_menu(self):
        if not self.isCompound():
            return []
        m = []
        if not self.get_gl_hl_use_array_idx():
            m.append(("Pick by component", self.onPickComponent, None))
        else:
            m.append(("Hide selected", self.onHidePickedComponent, None))
            m.append(("Show selected only", self.onShowPickedOnly, None))
            if len(self.hidden_component) > 0:
                m.append(("Show all", self.onShowAllComponent, None))
            m.append(("Pick as object", self.onUnpickComponent, None))

        return m

    def canvas_unselected(self):
        self._artists[0]._gl_hit_array_id = []

    def rect_contains(self, rect, check_selected_all_covered=False):
        ax = self.get_figaxes()._artists[0]
        a = self._artists[0]
        hit, all_covered, selected_idx = ax.gl_hit_test_rect(rect, a,
                                                             check_selected_all_covered=check_selected_all_covered)
        return hit, a, all_covered, selected_idx



from __future__ import print_function
#  Name   :fig_page
#
#          this class to manage matplotlib.figure
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************

from ifigure.matplotlib_mod.figure_mod import FigureMod as Figure
from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_axes import FigAxes
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import wx
import weakref
import ifigure.utils.cbook as cbook
import ifigure.events

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigPage')

frameart_zbase = -100000  # a big number so that it is draw first.


class FigPage(FigObj):
    def __init__(self, figsize=(0.15, 0.1), dpi=-1, src=None,
                 *args, **kywds):
        # list of page obj
        # this is for text, suptitle, and legend
        ###  placed in figure
        self.figobj = []
        super(FigPage, self).__init__()

        if dpi != -1:
            import warnings
            warnings.warn(
                "Screen DPI is set from system"
                "manually setting DPI is not supported in FigPage",
            UserWarning,
            stacklevel=2)

        #self.setp("dpi", dpi)

        # left, righ, top, bottom (fraction of rect)
        self.setp("def_margin", [0.15, 0.1, 0.15, 0.15])
        self.setp("figsize", figsize)

        self.setp("args", args)
        self.setp("kywds", kywds)
        self.setp("suptitle_labelinfo",  ['', 'black', 'default',
                                          'default', 'default', 'default'])

        self.setp("suptitle", '')
        self.setp("suptitle_size", 14)
        self.setp("title_size", 12)
        self.setp("ticklabel_size", 12)
        self.setp("axeslabel_size", 12)
        self.setp("axesbox_width",  1.0)
        self.setp("axestick_width",  1.0)
        self.setp("tick_font", 'sans-serif')
        self.setp("tick_weight", 'roman')
        self.setp("tick_style", 'normal')
        self.setp("title_font", 'sans-serif')
        self.setp("title_weight", 'roman')
        self.setp("title_style", 'normal')
        self.setp("nticks", [7, 7, 7])

        self._title_artist = None
        self._resize_cb = weakref.WeakSet()
        self._pan_cb = []
        self._zoom_cb = []
        self._nomargin = False
        self._page_margin = [0., 0., 0., 0.]  # left right bottom top?
        self._last_draw_size = (-1, -1)

    @classmethod
    def isFigPage(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'page'

    @classmethod
    def property_for_shell(self):
        return ["bgcolor", "page_title_size", "page_ticklabel_size", "page_axeslabel_size",
                "page_axesbox_width", "page_axestick_width", ]

    def can_have_child(self, child=None):
        from ifigure.mto.fig_book import FigBook
        if child is None:
            return True
        if isinstance(child, FigPage):
            return False
        if isinstance(child, FigBook):
            return False
        return True

    @classmethod  # define _attr values to be saved
    def property_in_file(self):
        return (['facecolor', 'figwidth', 'figheight'] +
                super(FigPage, self).property_in_file())

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ["tick_font", "tick_weight", "tick_style", "nticks",
                "title_font", "title_weight", "title_style",
                "ticklabel_size", "axeslabel_size",
                "def_margin", "axestick_width", "axesbox_width",
                "axestick_width", "suptitle_labelinfo", "figsize"]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'page.png')
        return [idx1]

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        if key == wx.WXK_RETURN:
            return self.onShow
        return None

    def realize(self):
        if not self._suppress:
            from ifigure.mto.fig_axes import FigAxes
            from ifigure.mto.figobj_gpholder import FigObjGPHolder
            self.generate_artist()
            axes = [o for o in self.walk_tree() if isinstance(o, FigAxes)]
            o1 = [o for o in self.walk_tree() if (not isinstance(o, FigAxes) and
                                                  not isinstance(o, FigObjGPHolder))]
            o2 = [o for o in self.walk_tree() if isinstance(o, FigObjGPHolder)]
            oo = [o for o in axes + o1 + o2 if o is not self]

            for o in oo:
                o.generate_artist()
        else:
            if not self.isempty():
                self.del_artist(delall=True)
            for objname, figobj in self.get_children():
                figobj.realize()

    def realize_children(self):
        for objname, figobj in self.get_children():
            figobj.realize(realize_gpholder='non_gp')

        for objname, figobj in self.get_children():
            figobj.realize(realize_gpholder='gp')

        #from ifigure.mto.figobj_gpholder import FigObjGPHolder

        # not_gp_holder = [figobj for objname, figobj in self.get_children()
        #                 if not isinstance(figobj, FigObjGPHolder)]

        # gp_holder = [figobj for objname, figobj in self.get_children()
        #             if isinstance(figobj, FigObjGPHolder)]

        # for figobj in not_gp_holder:
        #    figobj.realize()

        # for figobj in gp_holder:
        #    figobj.realize()

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        if self.isempty():
            dpi = wx.GetApp().get_dpi()
            dprint2("generatieng figure")
            artist = Figure(figsize=self.getp("figsize"),
                            dpi=dpi,
                            *(self.getp("args")),
                            **(self.getvar("kywds")))
            self._artists.append(artist)
            self._title_artist = artist.suptitle('')
            artist.figobj = self
            artist.figobj_hl = []
            if self.hasp("loaded_property"):
                lp = self.getp("loaded_property")
                if len(lp) > 0:
                    self.set_artist_property(self._artists[0], lp[0])
                self.delp("loaded_property")

            # set suptitle
            info = self.getp('suptitle_labelinfo')
            self.set_suptitle(info[0],
                              size=info[-1],
                              color=info[1])

    def del_artist(self, artist=None, delall=False):

        if delall:
            alist = self._artists
        else:
            alist = artist

        if self._title_artist is not None:
            #self._artists[0].texts.remove(self._title_artist)
            self._title_artist.remove()
            self._title_artist = None

        self.store_loaded_property()

        if len(alist) == len(self._artists):
            for name, child in self.get_children():
                child.del_artist(delall=True)

        super(FigPage, self).del_artist(alist, delall=delall)

    def add_axes(self, name=None, *args, **kywds):
        fig_axes = FigAxes(*args, **kywds)
        if name is None:
            name = self.get_next_name("axes")

        fig_axes.setp("margin", self.getp("def_margin"))
        self.add_child(name, fig_axes)
        return self.get_iaxes(fig_axes)

    def del_axes(self, fig_axes):
        fig_axes.destroy()

    def get_iaxes(self, ax):
        c = 0
        for name, child in self.get_children():
            if isinstance(child, FigAxes):
                if ax is child:
                    return c
                c = c + 1
        for child in self.walk_tree():
            if isinstance(child, FigAxes):
                if child._floating and child._generic_axes:
                    if ax is child:
                        return c
                    c = c + 1

    def get_axes(self, iax=None):
        # if iax is None: return all axes
        # in this case, returned list is
        # the same as get_children: (name, obj)
        c = 0
        if iax is None:
            return [x for x in self.get_children()
                    if isinstance(x[1], FigAxes)]

        for name, child in self.get_children():
            if isinstance(child, FigAxes):
                if c == iax:
                    return child
                c = c + 1
        for child in self.walk_tree():
            if isinstance(child, FigAxes):
                if child._floating and child._generic_axes:
                    if c == iax:
                        return child
                    c = c + 1
        return None

    def i_axes(self, ax):
        return self.get_iaxes(ax)

    def num_axes(self, include_floating=False):
        c = 0
        for name, child in self.get_children():
            if isinstance(child, FigAxes):
                c = c + 1
        if not include_floating:
            return c
        for child in self.walk_tree():
            if isinstance(child, FigAxes):
                if child._floating and child._generic_axes:
                    c = c + 1
        return c

    def walk_axes(self):
        c = 0
        for name, child in self.get_children():
            if isinstance(child, FigAxes):
                yield child

    def reset_axesbmp_update(self):
        figure = self._artists[0]
        for a in figure.axes:
            if (hasattr(a, 'figobj') and
                    a.figobj is not None):
                a.figobj.set_bmp_update(False)

    def set_bmp_update(self, value):
        # all axes will be updated
        # currently it is called when page default size are chagned.
        if len(self._artists) == 0:
            return
        if not value:
            self.reset_axesbmp_update()

    def set_area(self, areas):
        naxes = self.num_axes()
        fa = self.get_axes()
        narea = len(areas)
        if naxes > narea:
            for k in range(narea, naxes):
                dprint1('deleting '+str(fa[k][1]))
                self.del_axes(fa[k][1])

        i_r = 0
        for area in areas:
            if i_r < naxes:
                ax = fa[i_r][1]
                ax.set_area(area)
            else:
                iax = self.add_axes(area=area)
                ax = self.get_axes(iax)
                if self.isempty():
                    self.realize()
                ax.realize()
            i_r = i_r+1

    def get_area(self):
        return [o.getp("area") for o in self.walk_axes()]

    def set_def_margin(self, value, a=None):
        self.setp("def_margin", value[:])
        for f_axes in self.walk_axes():
            f_axes.set_area(f_axes.get_area())

    def get_def_margin(self, a=None):
        return self.getp("def_margin")[:]

    def set_page_margin(self, value, a=None):
        self._page_margin = value[:]
        for f_axes in self.walk_axes():
            f_axes.set_area(f_axes.get_area())

    def get_page_margin(self, a=None):
        return self._page_margin[:]

    def set_section(self, *args, **kargs):
        # utility to split a page
        def merge(a, b):
            def feq(a, b):
                th = 1e-5
                return abs(a-b) < th
            if (feq(a[3], b[3]) and feq(a[1], b[1])):  # horizontal merge
                if feq(a[0]+a[2], b[0]):
                    return [a[0], a[1], a[2]+b[2], a[3]]
                if feq(b[0]+b[2], a[0]):
                    return [b[0], a[1], a[2]+b[2], a[3]]
            if (feq(a[2], b[2]) and feq(a[0], b[0])):  # vertical merge
                if feq(a[1]+a[3], b[1]):
                    return [a[0], a[1], a[2], a[3]+b[3]]
                if feq(b[1]+b[3], a[1]):
                    return [a[0], b[1], a[2], a[3]+b[3]]
            return None

        def check_d(a, lim):
            if (not isinstance(a, list) and
                    not isinstance(a, tuple)):
                a = [a]
            if len(a) != lim-1:
                print("number of dx (or dy) should be nrow-1 (or nnoc-1)")
                return None
            if sum(a) > 1 or sum(a) < 0:
                print("dx (or dy) is too large or too small")
                return None
            a.append(1.-sum(a))
            return a
        nrow = 1
        ncol = 1
        asym = []
        if (isinstance(args[0], list) or
                isinstance(args[0], tuple)):
            ncol = len(args[0])
            nrow = args[0]
        else:
            if len(args) > 0:
                nrow = args[0]
            if len(args) > 1:
                ncol = args[1]
            if len(args) > 2:
                asym = args[2:]
            nrow = [nrow]*ncol

        if nrow[0] == 0:
            self.set_area([])
            return

        dx = [1./ncol]*ncol
        dy = [[1./nrow[k]]*nrow[k] for k in range(ncol)]

        if 'dx' in kargs:
            dx = check_d(kargs["dx"], ncol)
            if dx is None:
                return
        if 'dy' in kargs:
            dy = check_d(kargs["dy"], nrow[0])

        areas0 = []
        for i in range(ncol):
            for j in range(nrow[i]):
                if i == 0:
                    x0 = 0
                else:
                    x0 = sum(dx[:i])
                if j == 0:
                    y0 = 1. - dy[i][0]
                else:
                    y0 = 1. - sum(dy[i][:j+1])
                areas0.append([x0, y0, dx[i], dy[i][j]])

        flag = [1]*len(areas0)
        mareas0 = []

        for a in asym:
            if max(a) > len(flag)-1:
                print('incorrect index for merge')
                return
            if min(a) < 0:
                print('incorrect index for merge')
                return
            if flag[a[0]] == 0:
                print('area is already merged')
                return
            marea = areas0[a[0]]
            flag[a[0]] = 0
            for b in a[1:]:
                if flag[b] == 0:
                    print('area is already merged')
                    return
                flag[b] = 0
                marea = merge(marea, areas0[b])
                if marea is None:
                    print('failed to merge area')
                    return
            mareas0.append(marea)

        areas = [areas0[i] for i in range(len(areas0)) if flag[i] == 1]
        for m in mareas0:
            areas.append(m)

        self.set_area(areas)
        if 'sort' in kargs:
            if (kargs['sort'] == 'col' or
                kargs['sort'] == 'column' or
                    kargs['sort'] == 'c'):
                #                print(('self.sort_axes_col',self.sort_axes_col))
                self.sort_axes_col()
            elif (kargs['sort'] == 'row' or
                  kargs['sort'] == 'r'):
                self.sort_axes_row()

#
#   mene in tree viewer
#
    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        if self.num_axes() == 0:
            m = ('Add Axes',   self.onAddAxes, None)
        else:
            m = ('-Add Axes',   self.onAddAxes, None)
        if self.get_figbook().isOpen:
            me = ('Show Page',      self.onShow,    None)
        else:
            me = ('-Show Page',      self.onShow,    None)
        return [m, me,
                ('---',   None, None),
                ('Realize',   self.onRealize, None),
                ('---',   None, None)] + \
            super(FigPage, self).tree_viewer_menu()

    def onRealize(self, e):
        self.realize()
        ifigure.events.SendPVDrawRequest(self)
 #      self.onForceUpdate(e)

    def onAddAxes(self, e):
        self.add_axes()
        self.realize()
        self.set_area([[0, 0, 1, 1]])
        ifigure.events.SendPVAddFigobj(self)
        e.Skip()
#       self.onForceUpdate(e)

    def onShow(self, e):
        id = e.GetEventObject()
        ifigure.events.SendShowPageEvent(self, id)

#    def onDelete(self, e):
#        super(FigPage, self).onDelete(e)
#       return

#       app=self.get_root_parent().app

#       pbook=(self.get_root_parent()).pbook
#       if pbook is not self.get_parent():
        # easy case...
#          self.destroy()
#          e.Skip()
#       else:
        # show correct page to canvas..
#          npage, ipage=self.get_nipage()
#          cpage=self.get_parent().get_page(app.ipage)
#          print cpage
#          if cpage is self:
#             new_ipage = min([app.ipage, npage-2])
#             app.show_page(new_ipage)
#             self.destroy()
#          else:
#             self.destroy()
#             npage, ipage=cpage.get_nipage()
#             app.show_page(ipage)
#       e.Skip()

    def get_nipage(self):
        # return i_page and num_page of self
        npage = 0
        ipage = 0
        for name, child in self.get_parent().get_children():
            try:
                if child is self:
                    ipage = npage
                if child.isFigPage():
                    npage = npage+1
            except Exception:
                pass
        return npage, ipage

    def onResize(self, evt):
        if not self.isempty():
            self.setp('figsize', (self._artists[0].get_figwidth(),
                                  self._artists[0].get_figheight()))
        for obj in self._resize_cb:
            obj.onResize(evt)

    def onZoom(self, evt, figaxes):
        for ref in self._zoom_cb:
            if ref() is not None:
                ref().onZoom(evt, figaxes)

    def onPan(self, evt, figaxes):
        for ref in self._pan_cb:
            if ref() is not None:
                ref().onPan(evt, figaxes)

    def add_resize_cb(self, figobj):
        #        if figobj in [cb() for cb in self._resize_cb]: return
        self._resize_cb.add(figobj)

    def rm_resize_cb(self, figobj):
        if figobj in self._resize_cb:
            self._resize_cb.remove(figobj)
#        self._resize_cb = [x for x in self._resize_cb if x() is not figobj]

    def set_nomargin(self, value):
        from ifigure.mto.fig_axes import FigInsetAxes
        self._nomargin = value
        for figaxes in self.walk_axes():
            if not isinstance(figaxes, FigInsetAxes):
                figaxes.apply_nomargin_mode()
                figaxes.set_bmp_update(False)

    def get_nomargin(self):
        return self._nomargin

    def set_suptitle(self, txt, a=None, size=None, color=None):
        info = self.getp('suptitle_labelinfo')
        info[0] = txt
        if size is not None:
            info[-1] = size
        if color is not None:
            info[1] = color
        if self._title_artist is not None:
            self._title_artist.set_text(txt)
            if info[-1] != 'default':
                self._title_artist.set_size(info[-1])
            if info[1] != 'default':
                self._title_artist.set_color(info[1])

    def get_suptitle(self, a=None):
        info = self.getp('suptitle_labelinfo')
#        if self._title_artist is not None:
#            info[0] = self._title_artist.get_text()
        return info[0]

    def set_suptitle_size(self, s, a=None):
        info = self.getp('suptitle_labelinfo')
        info[5] = s
        if self._title_artist is not None:
            self._title_artist.set_size(float(s))

    def get_suptitle_size(self, a=None):
        info = self.getp('suptitle_labelinfo')
        return info[5]

    def _set_textprop(self, value, a):
        a.set_text(value[0])
        a.set_color(value[1])
        a.set_family(value[2])
        a.set_weight(value[3])
        a.set_style(value[4])
        a.set_size(float(value[5]))

    def _get_textprop(self, a):
        return [a.get_text(),
                a.get_color(),
                a.get_family()[0],
                a.get_weight(),
                a.get_style(),
                str(a.get_size())]

    def set_suptitle_labelinfo(self, value, a=None):
        tinfo = self.getp('suptitle_labelinfo')

        tfont = self.getp('title_font')
        tweight = self.getp('title_weight')
        tstyle = self.getp('title_style')
        tsize = self.getp('title_size')

        v2 = [x for x in value]
        if str(value[2]) == 'default':
            v2[2] = tfont
        if str(value[3]) == 'default':
            v2[3] = tweight
        if str(value[4]) == 'default':
            v2[4] = tstyle
        if str(value[5]) == 'default':
            v2[5] = tsize

        if self._title_artist is not None:
            self._set_textprop(v2, self._title_artist)
        self.setp('suptitle_labelinfo', value)

    def get_suptitle_labelinfo(self, a=None):
        return [x for x in self.getp('suptitle_labelinfo')]

    def postprocess_mpltext_edit(self):
        '''
        called after text is edited interactively
        this allows to load mpl text data to figobj
        '''
        info = self.getp('suptitle_labelinfo')
        if self._title_artist is not None:
            info[0] = self._title_artist.get_text()

    def sort_axes_col(self):
        '''
        sort is done by using left top conner
        '''
        l = [(int((a.getp('area')[0])*30),
              int((a.getp('area')[1] + a.getp('area')[3])*30),
              a) for a in self.walk_axes()]
        cols = [int((a.getp('area')[0])*30)
                for a in self.walk_axes()]
        idx = 0
        for col, c in enumerate(sorted(set(cols))):
            el = [item[2] for item in l if item[0] == c]
            l2 = [(int((a.getp('area')[1] + a.getp('area')[3])*30),
                   a) for a in el]
            for row, item in enumerate(reversed(sorted(l2))):
                id1 = self.i_child(item[1])
                self.move_child(id1, idx)
                idx = idx+1

    def sort_axes_row(self):
        '''
        sort is done by using left top conner
        '''
        l = [(int((a.getp('area')[0])*30),
              int((a.getp('area')[1] + a.getp('area')[3])*30),
              a) for a in self.walk_axes()]
        rows = [int((a.getp('area')[1] + a.getp('area')[3])*30)
                for a in self.walk_axes()]
        idx = 0
        for row, c in enumerate(reversed(sorted(set(rows)))):
            el = [item[2] for item in l if item[1] == c]
            l2 = [(int((a.getp('area')[0])*30),
                   a) for a in el]
            for col, item in enumerate(sorted(l2)):
                id1 = self.i_child(item[1])
                self.move_child(id1, idx)
                idx = idx+1


# save/load
#    def _do_save_data(self, fid=None):
#        print "saving fig_page data"


    def save_data2(self, data):
        # the first element is version code
        param = {"_page_margin": self._page_margin,
                 "_nomargin": self._nomargin}
        data['FigPage'] = (1, self.getp(), param)
        data = super(FigPage, self).save_data2(data)
        return data

    def load_data2(self, data):
        self.handle_loaded_figobj_data(data, 'FigPage')
        super(FigPage, self).load_data2(data)

    def set_figure_dpi(self, value):
        if len(self._artists) == 0:
            self.realize()
        self._artists[0].set_dpi(value)

    def get_figure_dpi(self):
        return self._artists[0].get_dpi()

    def set_viewer_samerange_mode(self, value, a):
        viewer = self.get_root_parent().app.find_bookviewer(self.get_figbook())
        viewer.set_use_samerange(value)

    def get_viewer_samerange_mode(self, a):
        viewer = self.get_root_parent().app.find_bookviewer(self.get_figbook())
        return viewer.get_use_samerange()

    def set_nticks(self, *args):
        '''
        set page default nticks
        '''
        changed = False
        for k, value in enumerate(args):
            if self.getp('nticks')[k] != value:
                self.getp('nticks')[k] = value
                changed = True
        if not changed:
            return
        for ax in self.walk_axes():
            ax.refresh_nticks()

    def convert_to_tex_style_text(self, mode=True):
        from ifigure.utils.cbook import tex_escape_equation
        if self._title_artist is None:
            return
        if mode:
            s = self._title_artist.get_text()
            self.setp('suptitle', s)
            self._title_artist.set_text(tex_escape_equation(str(s)))
        else:
            s = self.getp('suptitle')
            self._title_artist.set_text(s)

    def add_frameart(self, figobj):
        z = figobj.get_zorder()
        figobj.set_zorder(z + frameart_zbase)

    def rm_frameart(self, figobj):
        z = figobj.get_zorder()
        figobj.set_zorder(z - frameart_zbase)



import os
import wx
import ifigure
import numpy as np
import wx.lib.platebtn as platebtn
import wx.lib.buttons as buttons
import wx.lib.agw.buttonpanel as bp
import ifigure.utils.cbook as cbook
from ifigure.mto.fig_axes import FigInsetAxes
import ifigure.widgets.dialog as dialog

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty, UndoRedoFigobjProperty, UndoRedoFigobjMethod

from ifigure.utils.wx3to4 import image_GetAlpha, image_SetAlpha, image_SetOptionInt, evt_GetPosition, wxEmptyImage, wxStockCursor, wxCursorFromImage


def make_bitmap_with_bluebox(bitmap):
    # second bitmap has blue box around it
    h, w = bitmap.GetSize()
    image = bitmap.ConvertToImage()
    alpha = np.frombuffer(bytes(image_GetAlpha(image)),
                          dtype=np.uint8).reshape(w, h, -1)
    array = np.frombuffer(bytes(image.GetData()), dtype=np.uint8)
    array = array.copy()
    alpha = alpha.copy()

    array = array.reshape(w, h, -1)
    array[0, 1:-2, :1] = 0
    array[-1, 1:-2, :1] = 0
    array[1:-2, 0, :1] = 0
    array[1:-2, -1, :1] = 0

    array[1, 1, :1] = 0
    array[1, -2, :1] = 0
    array[-2, 1, :1] = 0
    array[-2, -2, :1] = 0

    array[1, 1, 2] = 255
    array[1, -2, 2] = 255
    array[-2, 1, 2] = 255
    array[-2, -2, 2] = 255

    array[0, 1:-2, 2] = 255
    array[-1, 1:-2, 2] = 255
    array[1:-2, 0, 2] = 255
    array[1:-2, -1, 2] = 255

    alpha[1:-2, -1] = 127
    alpha[1:-2, 0] = 127
    alpha[0, 1:-2] = 127
    alpha[-1, 1:-2] = 127
    alpha[1, 1] = 127
    alpha[1, -2] = 127
    alpha[-2, 1] = 127
    alpha[-2, -2] = 127
    image = wxEmptyImage(h, w)
    image.SetData(array.tobytes())
    image_SetAlpha(image, alpha)
    bitmap2 = image.ConvertToBitmap()

    return bitmap2


'''
button tasks
    ('pmode',           # task name
     'pmode.png',       # icon file
      1,                # toggle group id
     'plot mode',       # tips
      callable = None)  # method to be called when button is hit.
'''

three_d_bar = False
btasks0 = [  # ('previous', 'arrowleft.png', 0, 'previous page'),
    #        ('next',   'arrowright.png', 0,'next page'),
    #        ('---',    (10,10), 0, ''),
    #                   ('copy',   'copy.png', 0, ''),
    #                   ('paste',  'paste.png', 0, ''),
    ('pmode', 'pmode.png', 1, 'plot mode'),
    ('amode', 'amode.png', 1, 'annotation mode'),
    ('---', (10, 10), 0, ''), ]

btasks1_std2d_base = [('select', 'select.png', 1, 'select',),
                      ('zoom', 'zoom2.png', 1,
                       '\n'.join(['zoom', ' shift: zoom down', ' alt: menu to pick direction'])),
                      ('pan', 'arrowmove.png', 1, '\n'.join(
                          ['pan', ' shift: pan all']),),
                      ('cursor', 'cursor.png', 1, 'cursor',),
                      ('---', (10, 10), 0, ''), ]
btasks1_std2d = btasks1_std2d_base + [
    ('xlog', 'xlog.png', 0, 'toggle xlog',),
    ('ylog', 'ylog.png', 0, 'toggle ylog',),
    ('xauto', 'xauto.png', 0, 'autoscale x', ),
    ('yauto', 'yauto.png', 0, 'autoscale y', ),
    ('grid', 'grid.png', 0, 'toggle grid', ),
    ('nomargin', 'margin.png', 0, 'no margin mode', ), ]

btasks1_std3d_base = [('select', 'select.png', 1, 'select',),
                      ('zoom', 'zoom2.png', 1,
                       '\n'.join(['zoom', ' shift: zoom down', ' alt: menu to pick direction'])),
                      ('pan', 'arrowmove.png', 1, '\n'.join(
                          ['pan', ' shift: pan all']),),
                      ('cursor', 'cursor.png', 1, 'cursor',),
                      ('3dzoom', 'threed_rot.png', 1,
                       '3D zoom\n shift: pan\n alt: zoom',),
                      ('---', (10, 10), 0, ''), ]

btasks1_std3d = btasks1_std3d_base + [
    ('xauto', 'xauto.png', 0, 'autoscale x', ),
    ('yauto', 'yauto.png', 0, 'autoscale y', ),
    ('nomargin', 'margin.png', 0, 'no margin mode', ), ]

btasks2 = [  # ('selecta', 'select.bmp', 1, 'select'),
    ('text', 't.png', 1, 'insert text'),
    ('line', 'line.png', 1, 'insert line'),
    #                 ('curve2', 'curve.png', 1, 'insert curve'),
    ('curve', 'curve.png', 1, 'insert curve'),
    ('arrow', 'arrow.png', 1, 'insert arrow'),
    ('circle', 'circle.png', 1, 'insert circle'),
    ('rect', 'box.png', 1, 'insert rectangle'),
    ('legend', 'legend.png', 1, 'insert legend'),
    ('colorbar', 'colorbar.png', 1, 'insert colorbar'),
    ('eps', 'eps.png', 1, 'insert eps'), ]

btasks3 = [('copy_cb', 'copy.png', 0, 'copy image to clipboard'),
           ('save_pic', 'save_pic.png', 0, 'save graphic'),
           ('mail_pic', 'mail_pic.png', 0, 'mail graphic'),
           ('toggle_prop', 'form.png', 0, 'show property'),
           #         ('trash', 'trash.png', 0, 'delete page')
           ]

hotspot = {'zoom': [5, 5], 'circle': [8, 8], 'rect': [8, 8],
           'curve': [10, 10], 'curve2': [10, 10],
           'line': [2, 15], 'pan': [8, 8]}

zoom_crs = None
pan_crs = None


class TaskBtnList(list):
    def get_btn(self, name):
        for m in self:
            if m.btask == name:
                return m
        return


class FakeEvent(object):
    def GetEventObject(self):
        return self.obj

    def Skip(s):
        return


class ButtonInfo(bp.ButtonInfo):
    def __init__(self, *args, **kwargs):
        bp.ButtonInfo.__init__(self, *args, **kwargs)
        self.use_in_2d_menu = True
        self.use_in_3d_menu = True

    def GetBestSize(self):
        size = self.GetBitmap().GetSize()
        h = 26
        return (int((h - size[1]) / 2 + size[0]), int(h))


class ButtonPanel(bp.ButtonPanel):
    def DoGetBestSize(self):
        try:
            s = super(ButtonPanel, self).DoGetBestSize()
        except SystemError:
            #import traceback
            # traceback.print_stack()
            return 10, 10
        return s

    def Clear(self):
        # this is necessary for wxpython.2.9....
        self.Freeze()
        bp.ButtonPanel.Clear(self)
        # self.Thaw()

    def make_all_normal(self):
        for btn in self._vButtons:
            if not btn.IsEnabled():
                continue
            btn.SetStatus("Normal")


class navibar(ButtonPanel):
    def __init__(self, parent, id=-1, text='', *args, **kargs):
        super(navibar, self).__init__(parent, id, text, *args, **kargs)
        # mode of action (''=select, 'pan', 'zoom', 'text'....)
        self.mode = ''
        self.ptype = ''   # palette type ('pmode', 'amode')
        self.rotmode = False
        self.three_d_bar = False

        self.p1_btns = {'std2d': self.make_button_group(self, btasks1_std2d),
                        'std3d': self.make_button_group(self, btasks1_std3d), }

        self.p1_choice = ['std2d', 'std3d']
        self.p1_std_tasks = [btasks1_std2d[:],
                             btasks1_std3d[:], ]

        self._extra_buttons = {}          # store callbacks
        self._extra_buttons_refresh = {}  # store button refresh

        self.p0 = self.make_button_group(self, btasks0)
        self.p2 = self.make_button_group(self, btasks2)
        self.p3 = self.make_button_group(self, btasks3)

        self.SetPMode(skip_sc=True)

        self.zoom_up_down = 'up'
        self.zoom_menu = 0    # 0 or 1
        self.pan_all = 0
        self.Bind(wx.EVT_RIGHT_UP, self.OnRightUp)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
#        self.Bind(wx.EVT_KEY_UP,  self.OnKeyUp)

        from ifigure.ifigure_config import icondir
        from ifigure.utils.cbook import make_crs_list

        if zoom_crs is None:
            path1 = os.path.join(icondir, '16x16', 'zoom2.png')
            path2 = os.path.join(icondir, '16x16', 'zoom3minus.png')
            path3 = os.path.join(icondir, '16x16', 'zoom2menu.png')
            path4 = os.path.join(icondir, '16x16', 'zoom2minusmenu.png')
            globals()['zoom_crs'] = make_crs_list(
                [path1, path2, path3, path4], 5, 5)
            path1 = os.path.join(icondir, '16x16', 'arrowmove.png')
            path2 = os.path.join(icondir, '16x16', 'arrowmove_all.png')
            globals()['pan_crs'] = make_crs_list([path1, path2, ], 5, 5)
            path1 = os.path.join(icondir, '16x16', 'threed_rot.png')
            globals()['threed_crs'] = make_crs_list([path1, ], 5, 5)
        self.zoom_crs = zoom_crs
        self.pan_crs = pan_crs
        self.threed_crs = threed_crs
        self._curve_mode = 'pp'

    @property
    def p1(self):
        if self.three_d_bar:
            return self.p1_btns[self.p1_choice[1]]
        else:
            return self.p1_btns[self.p1_choice[0]]

    def install_palette(self, name, tasks0, mode='2D', refresh=None):
        '''
        tasks shou

        '''
        # we use taskname + '_' + name to avoid name conflict
        tasks = []
        for t in tasks0:
            if t[0] == '---':
                tasks.append(t)
            else:
                tasks.append([t[0] + '_' + name] + list(t[1:]))

        if mode == '2D':
            btasks = btasks1_std2d_base + list(tasks)
        else:
            btasks = btasks1_std3d_base + list(tasks)
        self.p1_btns[name] = self.make_button_group(self,
                                                    btasks)

        for t in tasks:
            self._extra_buttons[t[0]] = t[-1]
        if refresh is not None:
            self._extra_buttons_refresh[name] = refresh

        evt = FakeEvent()
        evt.obj = self

        tg = set([btnl.tg for btnl in self.p1_btns[name]
                  if isinstance(btnl, bp.ButtonInfo)])
        for t in tg:
            if t == 0:
                continue
            if t == 1:
                continue
            btnls = [b for b in self.p1_btns[name] if (isinstance(b, bp.ButtonInfo)
                                                       and b.tg == t)]
            btnls[0].SetBitmap(btnls[0].bitmap2)
            self._extra_buttons[btnls[0].btask](evt)
            for b in btnls[1:]:
                b.SetBitmap(b.bitmap1)

    def use_palette(self, name, mode='2D'):
        if mode == '2D':
            self.p1_choice[0] = name
        else:
            self.p1_choice[1] = name
        self.SetPMode()

    def use_std_palette(self):
        self.p1_choice = ['std2d', 'std3d']
        self.SetPMode()

    def refresh_palette(self):
        for name in self.p1_choice:
            if name in self._extra_buttons_refresh:
                self._extra_buttons_refresh[name](self,
                                                  self.p1_btns[name])

    def add_extra_group1_button(self, idx, data,
                                use_in_2d_menu=True,
                                use_in_3d_menu=False):
        xx = list(data[:4])
        if use_in_2d_menu:
            self.p1_std_tasks[0].insert(idx, xx)
            self.p1_btns['std2d'] = self.make_button_group(self,
                                                           self.p1_std_tasks[0])
        if use_in_3d_menu:
            self.p1_std_tasks[1].insert(idx, xx)
            self.p1_btns['std3d'] = self.make_button_group(self,
                                                           self.p1_std_tasks[1])

        self._extra_buttons[data[0]] = data[-1]
        self.SetPMode()

    def make_button_group(self, parent, btasks):
        from ifigure.ifigure_config import icondir

        bts = TaskBtnList()
        for items in btasks:
            btask, icon, tg, hint = items[:4]
            if btask == '---':
                bts.append('---')
#              bts.AddSpacer(icon)
                continue
            path = icon if os.path.isabs(icon) else os.path.join(icondir,
                                                                 '16x16', icon)
            if icon[-3:] == 'png':
                im = wx.Image(path, wx.BITMAP_TYPE_PNG)
                image = im.ConvertToBitmap()
#              im = im.ConvertToGreyscale()
#              im = im.ConvertToMono(0,0,0)
                if im.HasAlpha():
                    im.ConvertAlphaToMask()
                if btask in hotspot:
                    image_SetOptionInt(
                        im, wx.IMAGE_OPTION_CUR_HOTSPOT_X, hotspot[btask][0])
                    image_SetOptionInt(
                        im, wx.IMAGE_OPTION_CUR_HOTSPOT_Y, hotspot[btask][1])
                crs = wxCursorFromImage(im)
            elif icon[-3:] == 'bmp':
                im = wx.Image(path, wx.BITMAP_TYPE_BMP)
                image = im.ConvertToBitmap()
                if im.HasAlpha():
                    im.ConvertAlphaToMask()
                crs = wxCursorFromImage(im)

            btnl = ButtonInfo(self, wx.ID_ANY, image)
            btnl.custom_cursor = crs
            btnl.btask = btask
            btnl.bitmap1 = image
            btnl.bitmap2 = make_bitmap_with_bluebox(image)
            btnl.tg = tg
            if hint != '':
                btnl.SetShortHelp(hint)
            if tg > 0:
                btnl.SetKind('toggle')
            bts.append(btnl)
            #           def func(evt, btask=btask): return self.OnButton(evt, btask)
            #           parent.Bind(wx.EVT_BUTTON, func, btnl)
        return bts

#    def OnKeyUp(self, evt):
#       if evt.GetKeyCode() == wx.WXK_SHIFT:
#           if self.mode == 'zoom': self.SetZoomUpDown('Up')

    def OnKeyDown(self, evt):
        if evt.GetKeyCode() == wx.WXK_SHIFT:
            if self.mode == 'zoom':
                self.ToggleZoomUpDown()
            elif self.mode == 'pan':
                self.TogglePanAll()
        elif evt.GetKeyCode() == wx.WXK_ALT:
            if self.mode == 'zoom':
                self.ToggleZoomMenu()

    def OnRightUp(self, evt):
        ret = self.HitTest(evt_GetPosition(evt))
        if ret[0] == wx.NOT_FOUND:
            return super(navibar, self).OnLeftUp(evt)

        super(navibar, self).OnLeftUp(evt)
        btask = self.allbinfo[ret[0]].btask
        self.OnRightButton(evt, btask)

    def OnRightButton(self, evt, btask):
        if btask == 'mail_pic':
            self.MailPicRight(evt)
        elif btask == 'curve':
            self.CurveRight(evt)
        elif btask == 'grid':
            self.ToggleGridRight(evt)

    def OnLeftUp(self, evt):
        ret = self.HitTest(evt_GetPosition(evt))
        if ret[0] == wx.NOT_FOUND:
            return super(navibar, self).OnLeftUp(evt)

        super(navibar, self).OnLeftUp(evt)
        btask = self.allbinfo[ret[0]].btask
        self.OnButton(evt, btask)

    def OnButton(self, evt, btask):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)
        
        frame = cbook.FindFrame(self)
        btnl = [x for x in self.allbinfo if x.btask == btask][0]
        #        btnl = evt.GetEventObject()
        evt.SetEventObject(self)

        if btask == 'trash':
            frame.onDelPage(evt)
        elif btask == 'pmode':
            self.ExitInsertMode()
            self.SetPMode()
        elif btask == 'amode':
            self.ExitInsertMode()
            self.SetAMode()
        elif btask == 'select':
            self.SetSelect()
        elif btask == 'zoom':
            self.SetZoom(btnl)
        elif btask == 'pan':
            self.SetPan(btnl)
        elif btask == 'cursor':
            self.SetCursor()
        elif btask == '3dzoom':
            self.Set3DZoom(btnl)
        elif btask == 'next':
            frame.onNextPage(evt)
        elif btask == 'previous':
            frame.onPrevPage(evt)
        elif btask == 'xlog':
            self.SetXlog()
        elif btask == 'ylog':
            self.SetYlog()
        elif btask == 'xauto':
            self.SetXAuto()
        elif btask == 'yauto':
            self.SetYAuto()
        elif btask == 'save_pic':
            self.SavePic()
        elif btask == 'mail_pic':
            self.MailPic()
        elif btask == 'toggle_prop':
            self.ToggleProp()
        elif btask == 'copy_cb':
            self.CopyToCB()
#        elif btask == 'colorbar': self.ToggleCB()
        elif btask == 'grid':
            self.ToggleGrid()
        elif btask == 'nomargin':
            self.ToggleNoMargin()
        elif btask == '3d':
            self.Toggle3D()
        elif btask in self._extra_buttons:
            self.onHitExtra(evt, btnl)
        elif btask == 'line':
            self.GetParent().exit_layout_mode()  # just in case...
            if btnl.GetToggled():
                self._set_bitmap2(self.p2, self.p2.index(btnl))
                self.LineInsert(btask, btnl)
                self.SetInsertMode(btask, btnl)
            else:
                self.ExitInsertMode()
                #               self.SetAMode()
        elif btask == 'curve' and self._curve_mode == 'pp':
            self.GetParent().exit_layout_mode()  # just in case...
            if btnl.GetToggled():
                self._set_bitmap2(self.p2, self.p2.index(btnl))
                self.Curve2Insert(btask, btnl)
                self.SetInsertMode(btask, btnl)
            else:
                self.ExitInsertMode()
                #               self.SetAMode()
        else:
            self.GetParent().exit_layout_mode()  # just in case...
            if btnl.GetToggled():
                self._set_bitmap2(self.p2, self.p2.index(btnl))
                self.mode = btask
                self.SetInsertMode(btask, btnl)
            else:
                self.ExitInsertMode()
                #              self.SetAMode()

    def SetInsertMode(self, btask, btnl):
        for p in self.p2:
            if p != btnl:
                p.SetToggled(False)

        p = btnl
        sc = wxStockCursor(wx.CURSOR_WAIT)
        self.GetParent().canvas.SetCursor(p.custom_cursor)
#        print 'changing cursor'
        #

    def ExitInsertMode(self):
        #        print 'exiting insert mode'
        sc = wxStockCursor(wx.CURSOR_DEFAULT)
        self.GetParent().canvas.SetCursor(sc)
        self.mode = ''
        self.SetAMode()
        self._set_bitmap2(self.p2, -1)
        self.ptype = 'pmode'

    @property
    def abutton(self):
        return self.p0.get_btn('amode')

    @property
    def pbutton(self):
        return self.p0.get_btn('pmode')

    @property
    def iabutton(self):
        return self.p0.index(self.p0.get_btn('amode'))

    @property
    def ipbutton(self):
        return self.p0.index(self.p0.get_btn('pmode'))

    def SetAMode(self):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)
        
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)
        self.Clear()
        self.pbutton.SetStatus('Normal')
        self.abutton.SetStatus('Toggled')
        self.pbutton.SetToggled(False)
        self.abutton.SetToggled(True)
        self._set_bitmap2(self.p0, self.iabutton)
        for p in self.p2:
            p.SetToggled(False)
            p.SetStatus('Normal')
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p2:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.ptype = 'amode'
        self.GetParent()._show_cursor = False
        sc = wxStockCursor(wx.CURSOR_DEFAULT)
        self.GetParent().canvas.SetCursor(sc)

    def SetPMode(self, skip_sc=False):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)
        
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)
        self.Clear()
        self.pbutton.SetStatus('Toggled')
        self.abutton.SetStatus('Normal')
        self.pbutton.SetToggled(True)
        self.abutton.SetToggled(False)
        self._set_bitmap2(self.p0, self.ipbutton)
        for b in self.p0:
            self.AddButtonOrS(b)

        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.SetSelect(skip_sc=skip_sc)
        self.ptype = 'pmode'
        self.GetParent()._show_cursor = False

    def SetSelect(self, skip_sc=False):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)
        
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)
        self.Clear()
        self._set_bitmap2(self.p1, 0)
        self._set_toggle(0)
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p1:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.mode = ''
        self.GetParent()._show_cursor = False
        if not skip_sc:
            sc = wxStockCursor(wx.CURSOR_DEFAULT)
            self.GetParent().canvas.SetCursor(sc)

    def _set_bitmap2(self, array, index):
        for p in array:
            if isinstance(p, str):
                continue
            if p.tg < 2:
                p.SetBitmap(p.bitmap1)
        if index > -1:
            array[index].SetBitmap(array[index].bitmap2)

    def GoZoom(self):
        btnl = [x for x in self.allbinfo if x.btask == 'zoom'][0]
        self.SetZoom(btnl)

    def GoPan(self):
        btnl = [x for x in self.allbinfo if x.btask == 'pan'][0]
        self.SetPan(btnl)

    def SetZoom(self, btnl):
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)

        self.Clear()
        self._set_bitmap2(self.p1, 1)
        self._set_toggle(1)
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p1:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.mode = 'zoom'
        self.GetParent()._show_cursor = False
        self.zoom_up_down = 'up'
        self.zoom_menu = 0
        self.GetParent().canvas.SetCursor(btnl.custom_cursor)
        self.GetParent().exit_layout_mode()  # just in case...
        asel = self.GetParent().axes_selection
        if asel is None:
            return
        a = asel()
        if a is None:
            return
        if a.figobj is None:
            return
        if a.figobj.get_3d():
            self.GetParent().set_3dzoom_mode(True, pan_btn=11,
                                             zoom_btn=1, rotate_btn=10)

    def SetPan(self, btnl):
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)
        self.Clear()
        self._set_bitmap2(self.p1, 2)
        self._set_toggle(2)
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p1:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.mode = 'pan'
        self.pan_all = 0
        self.GetParent()._show_cursor = False
        self.GetParent().canvas.SetCursor(btnl.custom_cursor)
        self.GetParent().exit_layout_mode()  # just in case...

        asel = self.GetParent().axes_selection
        if asel is None:
            return
        a = asel()
        if a is None:
            return
        if a.figobj is None:
            return
        if a.figobj.get_3d():
            self.GetParent().set_3dzoom_mode(True, pan_btn=1,
                                             zoom_btn=11, rotate_btn=10)

    def SetCursor(self):
        if self.rotmode:
            self.rotmode = False
            self.GetParent().set_3dzoom_mode(False)
        self.Clear()
        self._set_bitmap2(self.p1, 3)
        self._set_toggle(3)
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p1:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.mode = 'cursor'
        self.GetParent()._show_cursor = True
        sc = wxStockCursor(wx.CURSOR_DEFAULT)
        self.GetParent().canvas.SetCursor(sc)
        self.GetParent().exit_layout_mode()  # just in case...

    def Set3DZoom(self, btnl):
        self.Clear()
        self._set_bitmap2(self.p1, 4)
        self._set_toggle(4)
        for b in self.p0:
            self.AddButtonOrS(b)
        for b in self.p1:
            self.AddButtonOrS(b)
        self.AddSpacer()
        for b in self.p3:
            self.AddButtonOrS(b)
        self.DoLayout()
        self.rotmode = True
        self.GetParent().set_3dzoom_mode(True)
        self.GetParent().canvas.SetCursor(btnl.custom_cursor)
        self.GetParent().exit_layout_mode()  # just in case...
        self.mode = '3dzoom'
        self.GetParent()._3d_rot_mode = 0

    def Set3DZoomCursor(self):
        self.GetParent().canvas.SetCursor(self.threed_crs[0])

    def SetPanCursor(self):
        self.GetParent().canvas.SetCursor(self.pan_crs[0])

    def SetZoomCursor(self):
        self.GetParent().canvas.SetCursor(self.zoom_crs[0])

    def _set_toggle(self, idx):
        mmm = 6
        for i in range(mmm):
            if self.p1[i] == '---':
                return
            if idx != i:
                self.p1[i].SetStatus('Normal')
                self.p1[i].SetToggled(False)
            else:
                self.p1[i].SetStatus('Toggled')
                self.p1[i].SetToggled(True)

    def Clear(self):
        self.allbinfo = []
        super(navibar, self).Clear()

    def AddButtonOrS(self, b):
        if isinstance(b, bp.ButtonInfo):
            self.AddButton(b)
            self.allbinfo.append(b)
        else:
            self.AddSeparator()

    def Show3DMenu(self):
        if self.three_d_bar:
            return
        self.three_d_bar = True
        globals()['three_d_bar'] = True
        if self.ptype == 'pmode':
            self.SetPMode()

    def Hide3DMenu(self):
        if not self.three_d_bar:
            return
        self.three_d_bar = False
        globals()['three_d_bar'] = False
        if self.ptype == 'pmode':
            self.SetPMode()

    def SetXYlog(self, name):
        asel = self.GetParent().axes_selection
        if asel is None:
            return
        a = asel()
        if a is None:
            return

        fig_axes = a.figobj
        p = fig_axes.get_axis_param(name)
        value = p.get_rangeparam()

#        print m
        ov = str(value[3])
        if ov == 'log':
            scale = 'linear'
        if ov == str('linear'):
            scale = 'log'

        canvas = self.GetParent()
        requests = canvas.make_xyauto_request(fig_axes, name, scale=scale)
        requests[fig_axes] = fig_axes.compute_new_range(
            request=requests[fig_axes])
        canvas.send_range_action(requests, 'range')

    def SetXlog(self):
        self.SetXYlog('x')

    def SetYlog(self):
        self.SetXYlog('y')

    def SetXAuto(self):
        self.GetParent().set_xauto()

    def SetYAuto(self):
        self.GetParent().set_yauto()

    def SavePic(self):
        self.GetParent().save_pic()

    def MailPic(self):
        self.GetParent().mail_pic()

    def MailPicRight(self, evt):
        def mailpic_set_pdf(evt, self=self):
            self.GetParent()._mailpic_format = 'pdf'

        def mailpic_set_eps(evt, self=self):
            self.GetParent()._mailpic_format = 'eps'

        def mailpic_set_svg(evt, self=self):
            self.GetParent()._mailpic_format = 'svg'

        def mailpic_set_jpeg(evt, self=self):
            self.GetParent()._mailpic_format = 'jpeg'

        def mailpic_set_png(evt, self=self):
            self.GetParent()._mailpic_format = 'png'

        def mailpic_set_pdfall(evt, self=self):
            self.GetParent()._mailpic_format = 'pdfall'

        def mailpic_set_gifanim(evt, self=self):
            self.GetParent()._mailpic_format = 'gifanim'
        data = ['eps', 'pdf', 'svg', 'png', 'jpeg', 'pdfall', 'gifanim']
        idx = data.index(self.GetParent()._mailpic_format)
        m = wx.Menu()
        menus = [
            ['EPS', mailpic_set_eps, None],
            ['PDF', mailpic_set_pdf, None],
            ['SVG', mailpic_set_svg, None],
            ['PNG', mailpic_set_png, None],
            ['JPEG', mailpic_set_jpeg, None],
            ['Multipage PDF', mailpic_set_pdfall, None],
            ['Animated Gif', mailpic_set_gifanim, None], ]
        menus[idx][0] = '^' + menus[idx][0]

        cbook.BuildPopUpMenu(m, menus)
        x, y = evt_GetPosition(evt)
        self.PopupMenu(m, [x, y])
        m.Destroy()

    def CurveRight(self, evt):
        def curve_set_freehand(evt, self=self):
            self._curve_mode = 'fh'

        def curve_set_pickpoints(evt, self=self):
            self._curve_mode = 'pp'
        data = ['pp', 'fh']
        idx = data.index(self._curve_mode)
        m = wx.Menu()
        menus = [
            ['PickPoints', curve_set_pickpoints, None],
            ['Freehand', curve_set_freehand, None], ]

        menus[idx][0] = '^' + menus[idx][0]
        cbook.BuildPopUpMenu(m, menus)
        x, y = evt_GetPosition(evt)
        self.PopupMenu(m, [x, y])
        m.Destroy()

    def CopyToCB(self):
        '''
        Copy_to_Clipboard_mod copys the buffer data
        which does not have highlight drawn
        '''
        frame = self.GetTopLevelParent()
        frame.onCopyToCB(None)

    def ToggleGridRight(self, evt):
        asel = self.GetParent().axes_selection
        if asel is None:
            return
        a = asel()
        if a is None:
            return
        ax = a.figobj
        value = ax.getp('grid')

        def grid_x(evt, self=self):
            value[0] = not value[0]
            ax.set_grid(value)
            a.figobj.set_bmp_update(False)
            canvas = self.GetParent().canvas
            canvas.draw()

        def grid_y(evt, self=self):
            value[1] = not value[1]
            ax.set_grid(value)
            a.figobj.set_bmp_update(False)
            canvas = self.GetParent().canvas
            canvas.draw()

        m = wx.Menu()
        menus = [
            ['X', grid_x, None],
            ['Y', grid_y, None], ]
        cbook.BuildPopUpMenu(m, menus)
        x, y = evt_GetPosition(evt)
        self.PopupMenu(m, [x, y])
        m.Destroy()

    def ToggleGrid(self):
        asel = self.GetParent().axes_selection

        if asel is None:
            return
        a = asel()
        if a is None:
            return
        ax = a.figobj
        value = ax.getp('grid')
        value[0] = not value[0]
        value[1] = not value[1]
        ax.set_grid(value)
        a.figobj.set_bmp_update(False)
        canvas = self.GetParent().canvas
        canvas.draw()

    def ToggleNoMargin(self):
        figure = self.GetParent().get_figure()
        figure.figobj.set_nomargin(not figure.figobj.get_nomargin())
        canvas = self.GetParent().canvas
        property_editor = self.GetParent().property_editor
        canvas.draw()
        property_editor.update_panel()

    def Toggle3D(self):
        asel = self.GetParent().axes_selection
        if asel is None:
            return
        a = asel()
        if a is not None:
            figobj = a.figobj

            if figobj is None:
                return
            if a.figobj.get_3d():
                if sum([len(x._member) for x in figobj._zaxis]) != 0:
                    dialog.message(self, 'panel has 3D plot', 'error')
                    return
                figobj.set_3d(False)
            else:
                figobj.set_3d(True)
#           figobj.reset_artist()
            canvas = self.GetParent().canvas
            canvas.draw()
            self.GetParent().set_axes_selection(figobj._artists[0])

    def LineInsert(self, btask, btnl):
        self.mode = 'line'
        self.GetParent().line_insert()
        for p in self.p2:
            if p != btnl:
                p.SetToggled(False)
        self.GetParent().canvas.SetCursor(btnl.custom_cursor)

    def Curve2Insert(self, btask, btnl):
        self.mode = 'curve2'
        self.GetParent().line_insert(mode=1)
        for p in self.p2:
            if p != btnl:
                p.SetToggled(False)
        self.GetParent().canvas.SetCursor(btnl.custom_cursor)

    def ToggleCB(self):
        asel = self.GetParent().axes_selection
        psel = self.GetParent().selection

        if asel is None:
            return
        a = asel()
        if a is None:
            return

        artists = [p() for p in psel]
        ax = a.figobj

        if len(artists) == 0:
            if False in [p.has_cbar() for p in ax._caxis]:
                for k, p in enumerate(ax._caxis):
                    if not p.has_cbar():
                        p.show_cbar(ax, offset=-0.1 * k)
            else:
                for p in ax._caxis:
                    p.hide_cbar(ax)
        else:
            figobjs = [a.figobj for a in artists]
            for figobj in figobjs:
                if figobj is None:
                    continue
                if hasattr(figobj, 'has_cbar'):  # check if cbar user
                    if figobj.has_cbar():
                        figobj.hide_cbar()
                    else:
                        index = ax._caxis.index(figobj.get_caxisparam())
                        figobj.get_caxisparam().show_cbar(ax, offset=-0.1 * index)
        canvas = self.GetParent().canvas
        canvas.draw()
        ifigure.events.SendChangedEvent(ax, w=canvas)

    def ToggleProp(self):
        top = self.GetTopLevelParent()
        top.toggle_property()

    def ToggleZoomUpDown(self):
        self.zoom_up_down = 'up' if self.zoom_up_down == 'down' else 'down'
        self._set_zoomcxr()

    def ToggleZoomForward(self):
        # up -> donw if it is already down return False
        if self.zoom_up_down == 'up':
            self.ToggleZoomUpDown()
            return True
        else:
            return False

    def ToggleZoomBackward(self):
        # down -> up if it is already up return False
        if self.zoom_up_down == 'down':
            self.ToggleZoomUpDown()
            return True
        else:
            return False

    def ToggleZoomMenu(self):
        self.zoom_menu = 1 if self.zoom_menu == 0 else 0
        self._set_zoomcxr()

    def TogglePanAll(self):
        self.pan_all = 1 if self.pan_all == 0 else 0
        self._set_pancxr()

    def onHitExtra(self, evt, btnl):
        btask = btnl.btask
        if btnl.tg == 0:
            self._extra_buttons[btask](evt)
            return
        else:
            btnls = [b for b in self.p1 if (isinstance(b, bp.ButtonInfo)
                                            and b.tg == btnl.tg)]
            self._extra_buttons[btask](evt)
            for b in btnls:
                if b == btnl:
                    b.SetToggled(True)
                    b.SetBitmap(b.bitmap2)
                else:
                    b.SetToggled(False)
                    b.SetBitmap(b.bitmap1)

    def ClickP1Button(self, name):
        '''
        programatically click a button, whose
        task is name
        '''
        p1 = self.p1
        if self.three_d_bar:
            choice = self.p1_choice[1]
        else:
            choice = self.p1_choice[0]

        pt = None
        for info in self.p1:
            if isinstance(info, str):
                continue
            if (info.btask == name or
                    info.btask == name + '_' + choice):
                rect = info.GetRect()
                pt = wx.Point(int(rect[0] + rect[1] / 2.),
                              int(rect[1] + rect[3] / 2.))

        if pt is None:
            return
        evt1 = wx.MouseEvent(wx.wxEVT_LEFT_DOWN)
        evt2 = wx.MouseEvent(wx.wxEVT_LEFT_UP)
        evt1.SetEventObject(self)
        evt2.SetEventObject(self)
        evt1.SetPosition(pt)
        evt2.SetPosition(pt)
        self.ProcessEvent(evt1)
        self.ProcessEvent(evt2)
        self.Refresh()
        wx.CallLater(100, self.make_all_normal)

    def _set_zoomcxr(self):
        if self.zoom_up_down == 'up':
            self.GetParent().canvas.SetCursor(
                self.zoom_crs[0 + self.zoom_menu * 2])
        else:
            self.GetParent().canvas.SetCursor(
                self.zoom_crs[1 + self.zoom_menu * 2])

    def _set_pancxr(self):
        self.GetParent().canvas.SetCursor(self.pan_crs[self.pan_all])



proj = obj.get_root_parent()
wdir = proj.eval("wdir")

loki = proj.psetting.loki
rdir = os.path.basename(wdir)
runname = os.path.basename(proj.eval("filename"))[:-4]
rdir = str(os.path.join('py_genray', runname))

files = ['genray.nc', 'lh_raytracing.nc', 'lh_raytracing_krf001.nc']
for file in files:
    file_f = os.path.join(wdir, file)
    loki.call_method('onRec', file=file_f, rfile=file, rdir=rdir)

solver = obj.get_parent()
sol = solver.sol



from matplotlib.patches import Rectangle
import wx
import numpy as np
from matplotlib.patches import PathPatch


def draw_field(fpage):
    figure = fpage._artists[0]

    a = 1./(2**(0.5))
    b = 1.0
    d = 1./(11*b + 10 * a)

    step = (2*a + 2*b)*d
    xy1 = [((a+b)*d+step*x, 0) for x in range(5)]
    xy2 = [(step*x, (a+b)*d) for x in range(6)]

    for m in range(6):
        for xy in xy1:
            hlp = Rectangle((xy[0], xy[1]+m*step), b*d, b*d, facecolor='r',
                            figure=figure, transform=figure.transFigure)
            figure.patches.append(hlp)
            hlp = Rectangle((xy[1]+m*step, xy[0]), b*d, b*d, facecolor='k',
                            figure=figure, transform=figure.transFigure)
            figure.patches.append(hlp)
    viewer = wx.GetApp().TopWindow.find_bookviewer(fpage.get_figbook())
    viewer.draw()


def draw_octagon(fpage, xy, d):
    import matplotlib.path as mpath
    Path = mpath.Path

    theta = [(22.5 + 45.*x)*np.pi/180 for x in range(8)]
    pathdata = [(Path.MOVETO, (xy[0]+d*cos(theta[-1]),
                               xy[1]+d*sin(theta[-1])), 0)]
    for t in theta:
        pathdata.append((Path.LINETO,  (xy[0]+d*cos(t),
                                        xy[1]+d*sin(t)), 0))



import numpy as np
import wx


def spec_demo():
    '''
    the demo in matplotlib. But calls
    interactive.specgram
    '''
    from pylab import arange, sin, where, logical_and, randn, pi

    dt = 0.0005
    t = arange(0.0, 20.0, dt)
    s1 = sin(2*pi*100*t)
    s2 = 2*sin(2*pi*400*t)

    # create a transient "chirp"
    mask = where(logical_and(t > 10, t < 12), 1.0, 0.0)
    s2 = s2 * mask

    # add some noise into the mix
    nse = 0.01*randn(len(t))

    x = s1 + s2 + nse  # the signal
    NFFT = 1024       # the length of the windowing segments
    Fs = int(1.0/dt)  # the sampling frequency

    from ifigure.interactive import figure, spec, nsec, plot, isec, clog, hold

    figure()
    hold(True)
    nsec(2)
    isec(0)
    plot(t, x)
    isec(1)
    spec(t, x, NFFT=NFFT, noverlap=900)
    clog()


def specgram_demo():
    '''
    the demo in matplotlib. But calls
    interactive.specgram
    '''
    from pylab import arange, sin, where, logical_and, randn, pi

    dt = 0.0005
    t = arange(0.0, 20.0, dt)
    s1 = sin(2*pi*100*t)
    s2 = 2*sin(2*pi*400*t)

    # create a transient "chirp"
    mask = where(logical_and(t > 10, t < 12), 1.0, 0.0)
    s2 = s2 * mask

    # add some noise into the mix
    nse = 0.01*randn(len(t))

    x = s1 + s2 + nse  # the signal
    NFFT = 1024       # the length of the windowing segments
    Fs = int(1.0/dt)  # the sampling frequency

    from ifigure.interactive import figure, specgram, nsec, plot, isec, clog, hold

    figure()
    hold(True)
    nsec(2)
    isec(0)
    plot(t, x)
    isec(1)
    specgram(x, NFFT=NFFT, Fs=Fs, noverlap=900)
    clog()


def image_demo():
    from ifigure.interactive import image, figure
    x = np.linspace(-5, 5, 41)
    y = np.linspace(-5, 5, 41)
    X, Y = np.meshgrid(x, y)
    r = np.sqrt(X*X + Y*Y)
    z = np.cos(r/10.*6*3.14)*np.exp(-r/3)
    figure()
    image(x, y, z)


def image_video_demo():
    from ifigure.interactive import image, figure
    x = np.linspace(-5, 5, 41)
    y = np.linspace(-5, 5, 41)
    X, Y = np.meshgrid(x, y)

    z = []
    for shift in np.linspace(-2., 2, 10):
        r = np.sqrt((X-shift)**2 + Y*Y)
        d = np.cos(r/10.*6*3.14)*np.exp(-r/3)
        z.append(d[np.newaxis, :, :])
    z = np.vstack(z)
    from ifigure.widgets.video_viewer import VideoViewer
    v = figure(viewer=VideoViewer)
    v.image(x, y, z)


def triplot_demo():
    from ifigure.interactive import triplot
    v = np.linspace(0.1, 10, 20)
    t = np.linspace(0, 3.1415, 20)

    V, T = np.meshgrid(v, t)
    x = (V*np.cos(T)).flatten()
    y = (V*np.sin(T)).flatten()

    triplot(x, y)


def tripcolor_demo():
    from ifigure.interactive import tripcolor
    v = np.linspace(0.1, 10, 80)
    t = np.linspace(0, 3.1415, 80)

    V, T = np.meshgrid(v, t)
    x = (V*np.cos(T)).flatten()
    y = (V*np.sin(T)).flatten()
    r = np.sqrt(x*x + y*y)
    z = np.cos(r/10.*6*3.14)*np.exp(-r/3)

    tripcolor(x, y, z)


def tripcolor_phasor_demo():
    from ifigure.interactive import tripcolor
    import time
    v = np.linspace(0.1, 10, 80)
    t = np.linspace(0, 3.1415, 80)

    V, T = np.meshgrid(v, t)
    x = (V*np.cos(T)).flatten()
    y = (V*np.sin(T)).flatten()
    r = np.sqrt(x*x + y*y)
    z = np.exp(1j*r/10.*6*3.14)*np.exp(-r/3)

    p = tripcolor(x, y, z)

    # aviewer import must be here. When there
    # is no window, tripcolor above will create
    # a viewer and aviewer in ifigure.interactive
    # is set.
    from ifigure.interactive import aviewer

    phase = np.linspace(0, np.pi*2, 30)
    for ph in phase:
        p.set_phasor(ph)
        p.get_figaxes().set_bmp_update(False)
        aviewer.draw()
        time.sleep(1)


def tripcolor_phasor_demo2():
    '''
    Example using wave viewer
    '''
    from ifigure.interactive import figure

    v = np.linspace(0.1, 10, 80)
    t = np.linspace(0, 3.1415, 80)

    V, T = np.meshgrid(v, t)
    x = (V*np.cos(T)).flatten()
    y = (V*np.sin(T)).flatten()
    r = np.sqrt(x*x + y*y)
    z = np.exp(1j*r/10.*6*3.14)*np.exp(-r/3)

    from ifigure.widgets.wave_viewer import WaveViewer
    v = figure(viewer=WaveViewer)
    v.tripcolor(x, y, z)


def tricontour_demo(**kwargs):
    from ifigure.interactive import tricontour

    X = np.linspace(-5, 5, 40)
    Y = np.linspace(-5, 5, 40)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)
    tricontour(X, Y, Z, cmap='coolwarm', **kwargs)


def contour_demo(**kwargs):
    from ifigure.interactive import contour

    X = np.linspace(-5, 5, 40)
    Y = np.linspace(-5, 5, 40)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)
    contour(X, Y, Z, cmap='coolwarm', **kwargs)


def scatter_demo(**kwargs):
    from ifigure.interactive import scatter

    x = np.random.random_sample(100)
    y = np.random.random_sample(100)
    c = np.random.random_sample(100)
    s = np.random.random_sample(100)*100+10

    scatter(x, y, c=c, s=s, cmap='coolwarm')


def hist_demo(**kwargs):
    from ifigure.interactive import hist, nsec, isec

    x1 = np.random.random_sample(4000)
    x2 = np.random.random_sample(400)*3
    hist(x1)
    nsec(3)
    isec(1)
    hist([x1, x2])
    isec(2)
    hist(x1.reshape(1000, 4))


def surf_demo(**kwargs):
    from ifigure.interactive import surf, threed

    threed('on')
    X = np.linspace(-5, 5, 40)
    Y = np.linspace(-5, 5, 40)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)
    surf(X, Y, Z, cmap='coolwarm', shade=True, **kwargs)



from __future__ import print_function
import weakref
import os
import traceback
import wx
import wx.grid
from ifigure.utils.eval_node import EvalNode
from ifigure.utils.cbook import ImageFiles, Write2Main, BuildPopUpMenu
from distutils.version import LooseVersion
from ifigure.utils.cbook import text_repr
import ifigure.widgets.dialog as dialog

from ifigure.utils.wx3to4 import isWX3, GridTableBase, TextEntryDialog

isWX_before_2_9 = LooseVersion(wx.__version__) < LooseVersion("2.9")

font_h = None
font_w = None
font = None


def set_default_font():
    size = 12
    font = wx.Font(pointSize=size, family=wx.DEFAULT,
                   style=wx.NORMAL,  weight=wx.NORMAL,
                   faceName='Consolas')
    globals()['font_label'] = wx.Font(pointSize=size, family=wx.DEFAULT,
                                      style=wx.NORMAL,  weight=wx.BOLD,
                                      faceName='Consolas')
    dc = wx.ScreenDC()
    dc.SetFont(font)
    w, h = dc.GetTextExtent('A')
    globals()['font_h'] = h*1.5
    globals()['font_w'] = w
    globals()['font'] = font


def get_varlist(td):
    if hasattr(td, 'get_shownvar'):
        names = td.get_shownvar()
    else:
        names = td.get_varlist()
    return names


class _PropertyGrid(wx.grid.Grid):
    def __init__(self, *args, **kargs):
        wx.grid.Grid.__init__(self, *args, **kargs)
        self.Bind(wx.EVT_SIZE, self.onSizeEvent)
#         colWindow = self.GetGridColLabelWindow()
#         colWindow.Bind(wx.EVT_PAINT, self.OnColPaint)

    def onSizeEvent(self, evt):
        evt.Skip()
        self.ForceRefresh()

    def SetTable(self, object, *attributes):
        if isWX3:
            self.tableRef = weakref.ref(object)
        else:
            self.tableRef = object
        return wx.grid.Grid.SetTable(self, object, *attributes)

    def GetTable(self):
        if isWX3:
            return self.tableRef()
        else:
            return self.tableRef

    # def OnPaint(self, evt):
    #    print("Here")


class VarViewerGValue(object):
    '''
    varialbe which can be edit by 
    varviwer
    '''

    def __init__(self, var, note):
        self._var = var
        self._note = note
        self._allow_eval = False

    def get_varlist(self):
        return list(self._var)

    def hasvar(self, name):
        return name in self._var

    def setvar(self, *args):
        if len(args) == 2:
            self._var[args[0]] = args[1]
        if len(args) == 1:
            if isinstance(args[0], dict):
                self._var = args[0]

    def getvar(self, name=None):
        if name is None:
            return self._var
        try:
            return self._var[name]
        except KeyError:
            return None

    def setnote(self, *args):
        if len(args) == 2:
            try:
                a = self._var[args[0]]
            except KeyError:
                return
            self._note[args[0]] = args[1]
        if len(args) == 1:
            self._note = args[0]

    def getnote(self, name=None):
        if name is None:
            return self._note
        try:
            return self._note[name]
        except KeyError:
            return ''

    def delvar(self, name):
        try:
            del self._var[name]
        except KeyError:
            pass
        try:
            del self._note[name]
        except KeyError:
            pass

    def get_drag_text2(self, key):
        ''' 
        text for dnd from var viewer. should be
        implemented at derived class
        '''
        pass


class TextDT(wx.TextDropTarget):
    pass


class VarViewerGDropTarget(wx.TextDropTarget):
    # obj is proj_viewer
    def __init__(self, obj):
        wx.TextDropTarget.__init__(self)
        self.obj = obj

    def OnDropText(self, x, y, indata):
        print("drop target")
        print(type(indata))

        # this is ad-hoc....!
        app = self.obj.GetTopLevelParent()
        indata = app._text_clip
        app._text_clip = ''

        data = '='+str(indata)
        try:
            print(str(indata))
            obj = EvalNode(str(indata))
            # print obj
            if obj.isTreeDict():
                data = data + '[:]'
        except Exception:
            pass
        pv = self.obj
        vv = pv.varviewer

        row = vv.grid.YToRow(y + vv.grid.GetViewStart()[1] *
                             vv.grid.GetScrollPixelsPerUnit()[1])
        if row == -1:
            return
        #        row=vv.grid.YToRow(y)
        gt = vv.grid.GetTable()
        treedict = gt.GetTreeDict()
        if treedict is None:
            return
        if row == -1:
            name = 'link'
            p = 1
            name1 = name+str(p)
            while treedict.getvar(name1) is not None:
                p = 1+p
                name1 = name+str(p)
            treedict.setvar(name1, data)

        else:
            name = (get_varlist(treedict))[row]
            treedict.setvar(name, data)

        vv.fill_list(treedict)
        item = pv.dt1._prev_item
        if pv.dt1._prev_item is not None:
            pv.tree.SetItemDropHighlight(
                pv.dt1._prev_item, False)

        pv.tree.UnselectAll()
        if item is not None:
            pv.tree.SelectItem(item)


class VarViewerGPopUp(wx.Menu):
    def __init__(self, parent):
        super(VarViewerGPopUp, self).__init__()
        self.parent = parent
        row = self.parent._startrow2
        if row != -1:
            gt = self.parent.grid.GetTable()
            txt = ['*', '*']
            if gt.sort == 0:
                txt[0] = '^'
            elif gt.sort == 1:
                txt[1] = '^'
            menus = [('Add...',    self.onAdd, None),
                     ('Duplicate...', self.onDuplicate, None),
                     ('Delete', self.onDelete, None),
                     ('Rename...', self.onRename, None),
                     ('+Sort', None, None),
                     (txt[0] + 'Up',   self.onSortUp, None),
                     (txt[1] + 'Down', self.onSortDown, None),
                     ('!',    None, None)]

            obj = gt.GetTreeDict()
            name = (get_varlist(obj))[row]
            if hasattr(obj, '_local_vars'):
                if name in obj._local_vars:
                    menus.append(('^Ignore Global', self.onToggleLocal, None))
                else:
                    menus.append(('*Ignore Global', self.onToggleLocal, None))
        else:
            menus = [('Add...',    self.onAdd, None), ]
        BuildPopUpMenu(self, menus)

    def onSortUp(self, e):
        gt = self.parent.grid.GetTable()
        if gt.sort == 0:
            gt.sort = -1
        else:
            gt.sort = 0
        self.parent.Refresh()

    def onSortDown(self, e):
        gt = self.parent.grid.GetTable()
        if gt.sort == 1:
            gt.sort = -1
        else:
            gt.sort = 1
        self.parent.Refresh()

    def onDuplicate(self, e):
        row = self.parent._startrow2
        if row == -1:
            return
        gt = self.parent.grid.GetTable()
        obj = gt.GetTreeDict()
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)
        dlg = TextEntryDialog(self.parent.GetTopLevelParent(),
                              "Enter the name of variable", "Duplicate variable", name+"_duplicated")
        if dlg.ShowModal() == wx.ID_OK:
            new_name = str(dlg.GetValue())
            if new_name != name:
                var = obj.getvar(name)
                obj.setvar(new_name, var)
#               obj.delvar(name)
                gt.SetTreeDict(obj)
        dlg.Destroy()

    def onAdd(self, e):
        row = self.parent._startrow2
        dlg = TextEntryDialog(self.parent.GetTopLevelParent(),
                              "Enter the name of variable", "Add variable", "")
        if dlg.ShowModal() == wx.ID_OK:
            new_name = str(dlg.GetValue())
            gt = self.parent.grid.GetTable()
            obj = gt.GetTreeDict()
            if row == -1:
                obj.setvar(new_name, None)
                gt.SetTreeDict(obj)
            else:
                name = (get_varlist(obj))[row]
                name = gt.get_row_name(row)
                if name.count(new_name) == 0:
                    obj.setvar(new_name, None)
                    gt.SetTreeDict(obj)
                else:
                    print("variable "+new_name + " exist!")
        dlg.Destroy()

    def onDelete(self, e):
        row = self.parent._startrow2
        if row == -1:
            return

        gt = self.parent.grid.GetTable()
        obj = gt.GetTreeDict()
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)

        ret = dialog.message(parent=self.parent.GetTopLevelParent(),
                             message='Do you want to delete ' + name + '?',
                             title="Delete variable",
                             style=4)
        if ret != 'yes':
            return
        obj.delvar(name)
        if hasattr(obj, '_local_vars'):
            if name in obj._local_vars:
                obj._local_vars.remove(name)
        gt.SetTreeDict(obj)

    def onToggleLocal(self, e):
        row = self.parent._startrow2
        if row == -1:
            return
        gt = self.parent.grid.GetTable()
        obj = gt.GetTreeDict()
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)
        if name in obj._local_vars:
            obj._local_vars.remove(name)
        else:
            obj._local_vars.append(name)

    def onRename(self, e):
        row = self.parent._startrow2
        if row == -1:
            return
        gt = self.parent.grid.GetTable()
        obj = gt.GetTreeDict()
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)

        dlg = TextEntryDialog(self.parent.GetTopLevelParent(),
                              "Enter the name of variable", "Add variable", name+"_rename")
        if dlg.ShowModal() == wx.ID_OK:
            new_name = str(dlg.GetValue())
            if new_name != name:
                var = obj.getvar(name)
                obj.setvar(new_name, var)
                obj.delvar(name)
                gt.SetTreeDict(obj)
                if (hasattr(obj, '_local_vars') and
                        name in obj._local_vars):
                    obj._local_vars.remove(name)
                    obj._local_vars.append(new_name)
        dlg.Destroy()


class VarViewerGridTable(GridTableBase):
    def __init__(self, obj, grid):
        GridTableBase.__init__(self)

        class Tmp(object):
            pass
        t = Tmp()
        self._obj = weakref.ref(t)
        del t
        self.sort = -1
        self._grid = grid
        self.currentRows = 0
        self.currentCols = self.GetNumberCols()

        self.SetTreeDict(obj)

    def getGrid(self):
        return self._grid

    def SetTreeDict(self, obj):
        if obj is None:
            class Tmp(object):
                pass
            obj = Tmp()
            self._obj = weakref.ref(obj)
            del obj
        else:
            self._obj = weakref.ref(obj)
        self.ResetView()

    def GetTreeDict(self):
        return self._obj()

    def GetNumberRows(self):
        """Return the number of rows in the grid"""
        obj = self._obj()
        if obj is None:
            return 0
        return len(obj.getvar())

    def GetNumberCols(self):
        """Return the number of columns in the grid"""
        return 4

    def IsEmptyCell(self, row, col):
        """Return True if the cell is empty"""
        return False

    def GetTypeName(self, row, col):
        """Return the name of the data type of the value in the cell"""
        # return None
        return 'string'

    def GetValue(self, row, col):
        """Return the value of a cell"""
        obj = self._obj()
        if obj is None:
            return ""
        names = get_varlist(obj)
        name = self.get_row_name(row)
        if row >= len(names):
            return ''
        if col == 0:
            val = obj.getvar(name)
            txt = text_repr(val)
            return txt
        if col == 1:
            val = obj.getvar(name)
            try:
                if val.startswith('='):
                    return "expression"
            except:
                pass
            #typstr = str(type(val))
            #txt = typstr[5:]
            #txt = txt[:-1]
            return type(val).__name__
        if col == 2:
            val = obj.getvar(name)
            if hasattr(val, 'shape'):
                return str(val.shape)
            else:
                return ''
        if col == 3:
            txt = obj.getnote(name)
            from ifigure.mto.py_code import PyParam
            if isinstance(obj, PyParam):
                path = obj.eval_all_keys(path=True)
                if names[row] in path:
                    arr = path[name]
                    if arr[-1][1] is not obj:
                        txt = txt + '('+arr[-1][1].get_full_path() + ')'
            return txt

        # self._list_str.append(obj.get_full_path()+'.getvar("'+key+'")')
        # self._keys=varss.keys()

    def SetValue(self, row, col, value):
        """Set the value of a cell"""
        pass

    def GetColLabelValue(self, col):
        if col == 0:
            v = 'value'
        elif col == 1:
            v = 'type'
        elif col == 2:
            v = 'shape'
        elif col == 3:
            v = 'description'
#        v = v+ u'\u2206'
        return v

    def GetRowLabelValue(self, row):
        obj = self._obj()
        if obj is None:
            return ""
        names = get_varlist(obj)
        if row >= len(names):
            super(VarViewerGridTable, self).GetRowLabelValue(row)
        return self.get_row_name(row)

    def get_row_name(self, row):
        obj = self._obj()
        if obj is None:
            return ''
        names = get_varlist(obj)
        if row >= len(names):
            return ''
        if self.sort == -1:
            return names[row]
        elif self.sort == 0:
            return sorted(names)[row]
        elif self.sort == 1:
            return [x for x in reversed(sorted(names))][row]

    def GetAttr(self, row, col, someExtraParameter):
        attr = wx.grid.GridCellAttr()
        obj = self._obj()
        if obj is None or row >= len(get_varlist(obj)):
            super(VarViewerGridTable, self).GetAttr(row, col,
                                                    someExtraParameter)
        if (col == 1 or col == 0 or col == 2):
            attr.SetReadOnly(1)
        if obj is not None:
            name = self.get_row_name(row)
            from ifigure.mto.py_code import PyParam
            if isinstance(obj, PyParam):
                path = obj.eval_all_keys(path=True)
                if name in path:
                    arr = path[name]
                    if arr[-1][1] is not obj:
                        attr.SetBackgroundColour('#707070')
                if name in obj._local_vars:
                    attr.SetBackgroundColour('#ffff00')
        return attr

    def ResetView(self):
        """Trim/extend the control's rows and update all values"""
        grid = self.getGrid()
        grid.BeginBatch()
        for current, new, delmsg, addmsg in [
            (self.currentRows, self.GetNumberRows(
            ), wx.grid.GRIDTABLE_NOTIFY_ROWS_DELETED, wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED),
            (self.currentCols, self.GetNumberCols(
            ), wx.grid.GRIDTABLE_NOTIFY_COLS_DELETED, wx.grid.GRIDTABLE_NOTIFY_COLS_APPENDED),
        ]:
            if new < current:
                msg = wx.grid.GridTableMessage(self, delmsg, new, current-new)
                # for i in range(current-new):
                grid.ProcessTableMessage(msg)
                self.currentRows = new
            elif new > current:
                msg = wx.grid.GridTableMessage(self, addmsg,  new-current)
                grid.ProcessTableMessage(msg)
                self.currentRows = new
        self.UpdateValues()
        obj = self._obj()
        if obj is not None:
            if len(get_varlist(obj)) == 0:
                ml = 8
            else:
                ml = max([len(x) for x in get_varlist(obj)])
#           grid.SetRowLabelSize(ml*10) #10 is an ad-hoc number
            grid.SetRowLabelSize((ml+2)*(font_w)+10)
        grid.EndBatch()

        # print 'current row', self.currentRows

        # The scroll bars aren't resized (at least on windows)
        # Jiggling the size of the window rescales the scrollbars
        h, w = grid.GetSize()
        grid.SetSize((h+1, w))
        grid.SetSize((h, w))
        grid.ForceRefresh()

    def fit_col_width(self):
        grid = self.getGrid()
        w, h = grid.GetClientSize()
        d = (w - grid.GetRowLabelSize()-grid.GetColSize(1) - grid.GetColSize(0)
             - grid.GetColSize(2))
        grid.SetColSize(3, max([d, 80]))

    def UpdateValues(self):
        """Update all displayed values"""
        msg = wx.grid.GridTableMessage(self,
                                       wx.grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.getGrid().ProcessTableMessage(msg)
        self.getGrid().ClearSelection()

    def IssueDoDragDrop(self, row):
        obj = self._obj()
        if obj is None:
            return ""

        name = self.get_row_name(row)
        text = obj.get_drag_text2(name)

        app = self._grid.GetTopLevelParent()
        app._text_clip = text

        tdo = wx.TextDataObject(text)
        src = self._grid.GetGridRowLabelWindow()
        tds = wx.DropSource(src)
        tds.SetData(tdo)
        tds.DoDragDrop(True)


class VarViewerG(wx.Panel):
    def __init__(self, parent):
        set_default_font()
        wx.Panel.__init__(self, parent)
        st1 = wx.StaticText(self, label=' = ')
        self.ct1 = wx.TextCtrl(self, value="",
                               style=wx.TE_PROCESS_ENTER)
        self.bx1 = wx.CheckBox(self, -1, 'Expression', (10, 10))
        sizer0 = wx.BoxSizer(wx.HORIZONTAL)
        sizer0.Add(st1, 0, wx.ALL, 1)
        sizer0.Add(self.ct1, 1, wx.ALL | wx.EXPAND, 1)
        sizer0.Add(self.bx1, 0, wx.ALL, 1)
        self.grid = _PropertyGrid(self)

        self.grid.CreateGrid(3, 0)
        self.grid.SetDefaultCellFont(font)
        self.grid.SetLabelFont(globals()['font_label'])
        self.grid.SetColLabelSize(int(font_h))
        self.grid.SetDefaultRowSize(int(font_h), True)
        self.grid.EnableDragColSize(True)
        self.grid.SetTable(VarViewerGridTable(None, self.grid))
        self.ct1.SetDropTarget(TextDT())

        bottom = wx.BoxSizer(wx.HORIZONTAL)
        modes = ['copy', 'paste', 'trash']
        for mode in modes:
            if mode == 'text':
                self.tc_path = TextCtrlPath(self, wx.ID_ANY, '')
                self.tc_path.pv = self
#              self.tc_path.Bind(wx.EVT_KEY_DOWN, self.onKeyPressed, self.tc_path)
                bottom.Add(self.tc_path, 1, wx.EXPAND | wx.ALL, 1)
            else:
                from ifigure.ifigure_config import icondir
                path = os.path.join(icondir, '16x16', mode+'.png')
                image = wx.Image(path, wx.BITMAP_TYPE_PNG).ConvertToBitmap()
                bt = wx.BitmapButton(self, -1, image)
                bt.SetToolTip(wx.ToolTip(mode))
                if mode == 'trash':
                    bottom.AddStretchSpacer()
                    bottom.Add(bt, 0)
                else:
                    bottom.Add(bt, 0)

                def func(e, mode=mode): return self.onButton(e, mode)
                self.Bind(wx.EVT_BUTTON, func, bt)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        sizer.Add(sizer0, 0, wx.EXPAND, 0)
        sizer.Add(self.grid, 1, wx.EXPAND, 0)
        sizer.Add(bottom, 0, wx.EXPAND)

        if isWX_before_2_9:
            pass
        else:
            self.Bind(wx.grid.EVT_GRID_CELL_CHANGING, self.onEdit)
        self.ct1.Bind(wx.EVT_TEXT_ENTER, self.onEditValue)

        rowWindow = self.grid.GetGridRowLabelWindow()
        self._potentialDrag = False
        self._startrow = None  # param for drag'n'drop
        self._startrow2 = -1  # param for popup menu
        rowWindow.Bind(wx.EVT_LEFT_DOWN, self.OnDragGridLeftDown)
        rowWindow.Bind(wx.EVT_LEFT_UP, self.OnDragGridLeftUp)
        rowWindow.Bind(wx.EVT_RIGHT_UP, self.OnRightRelease)
        rowWindow.Bind(wx.EVT_RIGHT_DOWN, self.OnRightPress)
#        rowWindow.Bind(wx.EVT_MOTION, self.OnRightRelease)

        self.grid.Bind(wx.EVT_SIZE, self.onGridSize)
        self.Bind(wx.grid.EVT_GRID_COL_SIZE, self.onGridSize)
#        self.Bind(wx.grid.EVT_GRID_BEGIN_DRAG, self.OnBeginDrag)
        self.Bind(wx.EVT_MOUSE_CAPTURE_LOST, lambda x: None)
        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_CLICK, self.onGridLClick)
        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.onGridLDClick)

        # this allows to select row always...
        try:
            self.grid.SetSelectionMode(wx.grid.Grid.wxGridSelectRows)
        except:  # wxPython 4.1.0
            self.grid.SetSelectionMode(wx.grid.Grid.GridSelectRows)

    def onGridLClick(self, evt):
        row = evt.GetRow()
        self._startrow = row

        self.grid.SelectRow(row)
        # self.grid.SetSelectionMode(wx.grid.Grid.wxGridSelectCells)
        self.set_expression_field(row)
        evt.Skip()

    def onGridLDClick(self, evt):
        print('Grid DClick')
        evt.Skip()

    def onButton(self, evt, mode):
        from ifigure.ifigure_config import vv_scratch
        import ifigure.utils.pickle_wrapper as pickle

        idx = self.grid.GetSelectedRows()
        if len(idx) == 0 and mode != 'paste':
            return
        gt = self.grid.GetTable()
        obj = gt.GetTreeDict()

        if mode == 'copy':
            name = str(self.grid.GetRowLabelValue(idx[0]))
            names = [str(self.grid.GetRowLabelValue(d)) for d in idx]

            self.GetTopLevelParent().set_status_text(
                'Copy ' + ', '.join(names), timeout=3000)
            try:
                fid = open(vv_scratch, 'wb')
                #data = {name: obj.getvar(name)}
                data = {n: obj.getvar(n) for n in names}
                pickle.dump(data, fid)
                fid.close()
            except:
                dialog.showtraceback(parent=self,
                                     txt='Failed to copy',
                                     title='Failed to copy',
                                     traceback=traceback.format_exc())
        elif mode == 'paste':
            if not os.path.exists(vv_scratch):
                dialog.showtraceback(parent=self,
                                     txt='paste data does not exists',
                                     title='Failed to paste',
                                     traceback='')
                return
            fid = open(vv_scratch, 'rb')
            data = pickle.load(fid)
            fid.close()
            for key in data:
                if obj.hasvar(key):
                    dlg = wx.MessageDialog(None,
                                           'Do you want to overwrite '+key + '?',
                                           'Variable already existws',
                                           wx.OK | wx.CANCEL)
                    ret = dlg.ShowModal()
                    dlg.Destroy()
                    if ret != wx.ID_OK:
                        continue
                self.GetTopLevelParent().set_status_text('Paste tree variable', timeout=3000)
                obj.setvar(key, data[key])
                obj._var_changed = True
            self.update()
        elif mode == 'trash':
            name = str(self.grid.GetRowLabelValue(idx[0]))
            names = [str(self.grid.GetRowLabelValue(d)) for d in idx]
            dlg = wx.MessageDialog(None,
                                   'Do you want to delete ' +
                                   ', '.join(names) + '?',
                                   'Deleting variable',
                                   wx.OK | wx.CANCEL)
            ret = dlg.ShowModal()
            dlg.Destroy()
            if ret == wx.ID_OK:
                for n in names:
                    obj.delvar(n)
                obj._var_changed = True
                self.GetTopLevelParent().set_status_text('Deltete ' +
                                                         ', '.join(names), timeout=3000)
                wx.CallAfter(self.update)

#        text = obj.get_drag_text2(name)

    def onGridSize(self, evt):
        self.grid.GetTable().fit_col_width()
        self.grid.ForceRefresh()

    def fill_list(self, obj):  # obj = TREEDICT
        gt = self.grid.GetTable()
        gt.SetTreeDict(obj)
        self.ct1.SetValue('')

    def update(self):
        gt = self.grid.GetTable()
        cobj = gt.GetTreeDict()
        try:
            if hasattr(cobj, '_var_changed'):
                if cobj._var_changed:
                    #                   print 'Reset Viewer', cobj
                    gt.ResetView()
                cobj._var_changed = False
        except:
            import traceback
            traceback.print_exc()
            pass

    def onEdit(self, e):
        row = e.GetRow()
        col = e.GetCol()

        obj = self.grid.GetTable().GetTreeDict()
        names = get_varlist(obj)

        if col == 3:  # edit desecription
            if row < len(names):
                key = names[row]
#               print e.GetString()
#               print self.grid.GetCellValue(row, col)
                obj._note[key] = str(e.GetString())

        self.fill_list(obj)

    def onEditValue(self, e):
        row = self._startrow
        gt = self.grid.GetTable()
        obj = gt.GetTreeDict()
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)
        if self.bx1.GetValue():
            val = '='+str(self.ct1.GetValue())
        else:
            val = eval(str(self.ct1.GetValue()))
        obj.setvar(name, val)
        self.grid.ForceRefresh()

    def SetDropTarget(self, tg):
        self.grid.GetGridWindow().SetDropTarget(tg)
#       uncomment this makes crush on MacOSX
#        self.grid.GetGridRowLabelWindow().SetDropTarget(tg)

    def OnDragGridLeftDown(self, e):
        # this is to convert y to row. need to adjust
        # scroll.....
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
        if row == -1:
            return
        self._startrow = row
        self._potentialDrag = True
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Bind(wx.EVT_MOTION, self.OnDragGridMotion)

        e.Skip()

    def OnDragGridLeftUp(self, e):
        """We are not dragging anymore, so unset the potentialDrag flag"""
        self._potentialDrag = False
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Unbind(wx.EVT_MOTION)

        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
        # row=self.grid.YToRow(e.GetY())
        if row == -1:
            return
        self.set_expression_field(row)
        '''
        gt=self.grid.GetTable()
        obj=gt.GetTreeDict()
        name=(get_varlist(obj))[row]
        name = gt.get_row_name(row)
        val=obj.getvar(name)
        txt=str(val)
        if ((isinstance(val, str) or isinstance(val, unicode)) and
               not val.startswith('=')):
               txt='"'+txt+'"'

        if txt.startswith('='):
           self.ct1.SetValue(txt[1:])
           self.bx1.SetValue(True)
        else:
           self.ct1.SetValue(txt)
           self.bx1.SetValue(False)
        '''
        e.Skip()

    def OnRightPress(self, e):
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
#        row=self.grid.YToRow(e.GetY())
        self._startrow2 = row

    def OnRightRelease(self, e):
        row = self.grid.YToRow(e.GetY() + self.grid.GetViewStart()[1] *
                               self.grid.GetScrollPixelsPerUnit()[1])
#        row=self.grid.YToRow(e.GetY())

        if self._startrow2 != row:
            # press and release happend at different row
            self._startrow2 = row
            return
        m = VarViewerGPopUp(self)
        self.PopupMenu(m,
                       e.GetPosition())
        m.Destroy()

    def OnDragGridMotion(self, e):
        rowWindow = self.grid.GetGridRowLabelWindow()
        rowWindow.Unbind(wx.EVT_MOTION)
        self._potentialDrag = False
        if self._startrow is None:
            return

        #print("Start drag")
        gt = self.grid.GetTable()
        gt.IssueDoDragDrop(self._startrow)

        e.Skip()

    def setdroptarget(self, proj_viewer):
        dt2 = VarViewerGDropTarget(proj_viewer)
        self.SetDropTarget(dt2)

    def Copy(self, e=None):
        idx = self.grid.GetSelectedRows()
        if len(idx) == 0:
            return
        name = str(self.grid.GetRowLabelValue(idx[0]))
        gt = self.grid.GetTable()
        obj = gt.GetTreeDict()
        text = obj.get_drag_text2(name)

        from ifigure.utils.cbook import SetText2Clipboard
        SetText2Clipboard(text)

    def Paste(self, e=None):
        '''
        can not paste to shellvar viewer
        '''
        pass

    def set_expression_field(self, row):
        gt = self.grid.GetTable()
        obj = gt.GetTreeDict()
        names = get_varlist(obj)
        if len(names) < row:
            return
        name = (get_varlist(obj))[row]
        name = gt.get_row_name(row)
        val = obj.getvar(name)
        txt = str(val)

        from ifigure.utils.cbook import isstringlike
        if (isstringlike(val) and not val.startswith('=')):
            txt = '"'+txt+'"'

        if txt.startswith('='):
            self.ct1.SetValue(txt[1:])
            self.bx1.SetValue(True)
        else:
            self.ct1.SetValue(txt)
            self.bx1.SetValue(False)



'''
  a utility to use futurize

'''
import sys
import os
import shutil
import subprocess as sp
import ifigure
futurize = None

##
futurize1 = os.path.join(os.path.dirname(os.path.realpath(sys.executable)), 'futurize')
if os.path.exists(futurize1):
    futurize = futurize1

##    
futurize2 = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.__file__))), 'bin', 'futurize')
if os.path.exists(futurize2):
    futurize = futurize2
    
## if piScope is installed using the same prefix as futurize, this one find it
futurize3 = os.path.join(os.path.dirname(
                         os.path.dirname(
                         os.path.dirname(
                         os.path.dirname(                             
                         os.path.dirname(ifigure.__file__))))), 'bin', 'futurize')
if os.path.exists(futurize3):
    futurize = futurize3

# probably the best is to keep only this one??    
if futurize is None:    
    futurize=shutil.which("futurize")
    
def call_futurize(file=None, dryrun=False, verbose=False, unicode=True,
                  stage1=True, stage2=False, help=False):
    
    def make_com0():
        command = [futurize,]
        if not dryrun:
            command.append('-w')
        if verbose:
            command.append('-v')
        return command

    def run_com(command):
        p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
        out, err = p.communicate()
        return out, err
    
    command = make_com0()
    if help:
        command = [futurize, '--help']
        out, err = run_com(command)
        print(out.decode('utf-8'))
        print(err.decode('utf-8'))        
        return
    if stage1:
        command.append('--stage1')
        command.append(file)        
        print(command)
        out, err = run_com(command)
        print(out.decode('utf-8'))
        print(err.decode('utf-8'))        

    if stage2:
        command.append('--stage2')
        command.append(file)                
        print(command)
        out, err = run_com(command)
        print(out.decode('utf-8'))
        print(err.decode('utf-8'))        

class futurizer():
    def process_script(self, obj, dryrun=False, verbose=False, unicode=True,
                       stage1=True, stage2=False, help=False):
        
        assert (futurize is not None), "futurize is not found"
        if help:
            call_futurize(help=True)
            return
        file = obj.path2fullpath()
        call_futurize(file=file, dryrun=dryrun, verbose=verbose, unicode=unicode,
                      stage1=stage1, stage2=stage2, help=False)

    def process_tree(self, root, dryrun=False, verbose=False, unicode=True,
                     stage1=True, stage2=False, help=False):
        assert (futurize is not None), "futurize is not found"        
        if help:
            call_futurize(help=True)
            return

        from ifigure.mto.py_script import PyScript

        for obj in root.walk_tree():
            print(obj)
            if isinstance(obj, PyScript):
                self.process_script(obj,  dryrun=dryrun, verbose=verbose, unicode=unicode,
                      stage1=stage1, stage2=stage2, help=False)
                
                
    def process_proj(self, dryrun=False, verbose=False, unicode=True,
                     stage1=True, stage2=False, help=False):
        import wx
        proj = wx.GetApp().TopWindow.proj
        self.process_tree(proj,  dryrun=dryrun, verbose=verbose, unicode=unicode,
                          stage1=stage1, stage2=stage2, help=help)
        
    process_folder = process_tree
                     
                          
